# visualization/__init__.py

"""Sub-package including functions used for visualization.

This sub-package includes the following modules:

- `latex`: Latex visaul creation functions.
- `util`: General visualization utilities.
- `video`: Video-based visualization functions.
"""

import os
import sys

FILE_ROOT=os.path.abspath(os.path.dirname(__file__))
PYTHON_ROOT=os.path.abspath(os.path.dirname(FILE_ROOT))
SRC_ROOT=os.path.abspath(os.path.dirname(PYTHON_ROOT))
RENDERING_ROOT=os.path.abspath(os.path.join(SRC_ROOT, "cpp/rendering/build/"))

sys.path.append(PYTHON_ROOT)
sys.path.append(RENDERING_ROOT)

from .latex import *
from .util import *
from .video import *



# cvt/io.py

"""A suite of common input/output functions.

This module includes several functions for reading and writing
different types of data useful for computer vision applications.

This module contains the following functions:

- `read_cams_sfm(camera_path, extension)` - Reads an entire directory of camera files in SFM format.
- `read_cams_trajectory(log_file)` - Reads camera file in Trajectory File format.
- `read_extrinsics_tum(tum_file, key_frames)` - Reads extrinsic camera trajectories in TUM format [timestamp tx ty tz qx qy qz qw].
- `read_matrix(mat_file)` - Reads a single matrix of float values from a file.
- `read_mesh(mesh_file)` - Reads a mesh from a file.
- `read_cluster_list(filename)` - Reads a cluster list file encoding supporting camera viewpoints.
- `read_pfm(pfm_file)` - Reads a file in *.pfm format.
- `read_point_cloud(point_cloud_file)` - Reads a point cloud from a file.
- `read_single_cam_sfm(cam_file, depth_planes)` - Reads a single camera file in SFM format.
- `read_stereo_intrinsics_yaml(intrinsics_file)` - Reads intrinsics information for a stereo camera pair from a *.yaml file.
- `write_cam_sfm()` -
- `write_matrix(M, mat_file)` - Writes a single matrix to a file.
- `write_mesh(mesh_file, mesh)` - Writes a mesh to a file.
- `write_pfm(pfm_file, data_map, scale)` - Writes a data map to a file in *.pfm format.
"""

import os
import sys
import numpy as np
import cv2
import re
import math
import open3d as o3d
import torch
from typing import List, Tuple

from scipy.spatial.transform import Rotation as rot

from camera import y_axis_rotation


def read_cams_sfm(camera_path: str, extension: str = "cam.txt") -> np.ndarray:
    """Reads an entire directory of camera files in SFM format.

    Parameters:
        camera_path: Path to the directory of camera files.
        extension: File extension being used for the camera files.

    Returns:
        Array of camera extrinsics, intrinsics, and view metadata (Nx2x4x4).
    """
    cam_files = os.listdir(camera_path)
    cam_files.sort()

    cams = []
    
    for cf in cam_files:
        if (cf[-7:] != extension):
            continue

        cam_path = os.path.join(camera_path,cf)
        #with open(cam_path,'r') as f:
        cam = read_single_cam_sfm(cam_path, 256)
        cams.append(cam)

    return np.asarray(cams)

def read_cams_trajectory(log_file: str) -> np.ndarray:
    """Reads camera file in Trajectory File format.

    Parameters:
        log_file: Input *.log file to be read.

    Returns:
        Array of camera extrinsics, intrinsics, and view metadata (Nx2x4x4).
    """
    cams = []
    
    with open(log_file,'r') as f:
        lines = f.readlines()

        for i in range(0,len(lines),5):
            cam = np.zeros((4, 4))
            # read extrinsic
            for j in range(1, 5):
                cam[j-1,:] = np.array([float(l.strip()) for l in lines[i+j].split()])
            cam = np.linalg.inv(cam)
                
            cams.append(cam)

    return cams

def read_extrinsics_tum(tum_file: str, key_frames: List[int] = None) -> np.ndarray:
    """Reads extrinsic camera trajectories in TUM format [timestamp tx ty tz qx qy qz qw].

    Parameters:
        tum_file: Input extrinsics file.
        key_frames: Indices corresponding to the desired keyframes.

    Returns:
        Array of camera extrinsics (Nx4x4).
    """
    rot_interval = 30
    max_rot_angle = math.pi / 3

    extrinsics = []
    with open(tum_file,"r") as tf:
        lines = tf.readlines()
        
        for i,line in enumerate(lines):
            l = np.asarray(line.strip().split(" "), dtype=float)
            l = l[1:]
            t = l[:3]
            q = l[3:]

            R = rot.from_quat(q).as_matrix()
            R = R.transpose()
            t = -R@t
            P = np.zeros((4,4))
            P[:3,:3] = R
            P[:3,3] = t.transpose()
            P[3,3] = 1

            extrinsics.append(P)

            if((key_frames == None) or (i in key_frames)):
                left = np.linspace(0.0, max_rot_angle, rot_interval)
                right = np.linspace(max_rot_angle, -(max_rot_angle), rot_interval*2)
                center = np.linspace(-(max_rot_angle), 0.0, rot_interval)
                thetas = np.concatenate((left,right,center))

                for theta in thetas:
                    new_P = y_axis_rotation(P,theta)
                    extrinsics.append(new_P)

    return np.asarray(extrinsics)

def read_matrix(mat_file: str) -> np.ndarray:
    """Reads a single matrix of float values from a file.

    Parameters:
        mat_file: Input file for the matrix to be read.

    Returns:
        The matrix stored in the given file.
    """
    with open(mat_file, 'r') as f:
        lines = f.readlines()
        M = []

        for l in lines:
            row = l.split()
            row = [float(s) for s in row]
            M.append(row)
        M = np.array(M)

    return M

def read_mesh(mesh_file: str) -> o3d.geometry.TriangleMesh:
    """Reads a mesh from a file.

    Parameters:
        mesh_file: Input mesh file.

    Returns:
        The mesh stored in the given file.
    """
    return o3d.io.read_triangle_mesh(mesh_file)

def read_cluster_list(filename: str) -> List[Tuple[int,List[int]]]:
    """Reads a cluster list file encoding supporting camera viewpoints.

    Parameters:
        filename: Input file encoding per-camera viewpoints.

    Returns:
        An array of tuples encoding (ref_view, [src_1,src_2,..])
    """
    data = []
    with open(filename) as f:
        num_views = int(f.readline())
        all_views = list(range(0,num_views))

        for view_idx in range(num_views):
            ref_view = int(f.readline().rstrip())
            src_views = [int(x) for x in f.readline().rstrip().split()[1::2]]
            if len(src_views) == 0:
                continue
            data.append((ref_view, src_views))
    return data

def read_pfm(pfm_file: str) -> np.ndarray:
    """Reads a file in *.pfm format.

    Parameters:
        pfm_file: Input *.pfm file to be read.

    Returns:
        Data map that was stored in the *.pfm file.
    """
    with open(pfm_file, 'rb') as pfm_file:
        color = None
        width = None
        height = None
        scale = None
        data_type = None
        header = pfm_file.readline().decode('iso8859_15').rstrip()

        if header == 'PF':
            color = True
        elif header == 'Pf':
            color = False
        else:
            raise Exception('Not a PFM file.')
        dim_match = re.match(r'^(\d+)\s(\d+)\s$', pfm_file.readline().decode('iso8859_15'))
        if dim_match:
            width, height = map(int, dim_match.groups())
        else:
            raise Exception('Malformed PFM header.')
        # scale = float(file.readline().rstrip())
        scale = float((pfm_file.readline()).decode('iso8859_15').rstrip())
        if scale < 0: # little-endian
            data_type = '<f'
        else:
            data_type = '>f' # big-endian
        data_string = pfm_file.read()
        data = np.fromstring(data_string, data_type)
        shape = (height, width, 3) if color else (height, width)
        data = np.reshape(data, shape)
        data = cv2.flip(data, 0)
    return data

def read_point_cloud(point_cloud_file: str) -> o3d.geometry.PointCloud:
    """Reads a point cloud from a file.

    Parameters:
        point_cloud_file: Input point cloud file.

    Returns:
        The point cloud stored in the given file.
    """
    return o3d.io.read_point_cloud(point_cloud_file)

def read_single_cam_sfm(cam_file: str, depth_planes: int = 256) -> np.ndarray:
    """Reads a single camera file in SFM format.

    Parameters:
        cam_file: Input camera file to be read.
        depth_planes: Number of depth planes to store in the view metadata.

    Returns:
        Camera extrinsics, intrinsics, and view metadata (2x4x4).
    """
    cam = np.zeros((2, 4, 4))

    with open(cam_file, 'r') as cam_file:
        words = cam_file.read().split()

    words_len = len(words)

    # read extrinsic
    for i in range(0, 4):
        for j in range(0, 4):
            extrinsic_index = 4 * i + j + 1
            cam[0,i,j] = float(words[extrinsic_index])

    # read intrinsic
    for i in range(0, 3):
        for j in range(0, 3):
            intrinsic_index = 3 * i + j + 18
            cam[1,i,j] = float(words[intrinsic_index])

    if words_len == 29:
        cam[1,3,0] = float(words[27])
        cam[1,3,1] = float(words[28])
        cam[1,3,2] = depth_planes
        cam[1,3,3] = cam[1][3][0] + (cam[1][3][1] * cam[1][3][2])
    elif words_len == 30:
        cam[1,3,0] = float(words[27])
        cam[1,3,1] = float(words[28])
        cam[1,3,2] = float(words[29])
        cam[1,3,3] = cam[1][3][0] + (cam[1][3][1] * cam[1][3][2])
    elif words_len == 31:
        cam[1,3,0] = words[27]
        cam[1,3,1] = float(words[28])
        cam[1,3,2] = float(words[29])
        cam[1,3,3] = float(words[30])
    else:
        cam[1,3,0] = 0
        cam[1,3,1] = 0
        cam[1,3,2] = 0
        cam[1,3,3] = 1

    return cam

def read_stereo_intrinsics_yaml(intrinsics_file: str) -> Tuple[ np.ndarray, \
                                                                np.ndarray, \
                                                                np.ndarray, \
                                                                np.ndarray, \
                                                                np.ndarray, \
                                                                np.ndarray]:
    """Reads intrinsics information for a stereo camera pair from a *.yaml file.

    Parameters:
        intrinsics_file: Input *.yaml file storing the intrinsics information.

    Returns:
        K_left: Intrinsics matrix (3x3) of left camera.
        D_left: Distortion coefficients vector (1x4) of left camera.
        K_right: Intrinsics matrix (3x3) of right camera.
        D_right: Distortion coefficients vector (1x4) of right camera.
        R: Relative rotation matrix (3x3) from left -> right cameras.
        T: Relative translation vector (1x3) from left -> right cameras.
    """
    K_left = np.zeros((3,3))
    D_left = np.zeros((1,4))
    K_right = np.zeros((3,3))
    D_right = np.zeros((1,4))
    R = np.zeros((3,3))
    T = np.zeros((1,3))

    cv_file = cv2.FileStorage(intrinsics_file, cv2.FILE_STORAGE_READ)

    left = cv_file.getNode("left")
    K_left = left.getNode("K").mat()
    D_left = left.getNode("D").mat()

    right = cv_file.getNode("right")
    K_right = right.getNode("K").mat()
    D_right = right.getNode("D").mat()

    R = cv_file.getNode("R").mat()
    T = cv_file.getNode("T").mat()

    cv_file.release()

    return [K_left, D_left, K_right, D_right, R, T]

def write_cam_sfm(cam_file: str, intrinsics: np.ndarray, extrinsics: np.ndarray) -> None:
    """Writes intrinsic and extrinsic camera parameters to a file in sfm format.

    Parameters:
        cam_file: The file to be writen to.
        intrinsics: Camera intrinsic data to be written.
        extrinsics: Camera extrinsic data to be written.
    """
    with open(cam_file, "w") as f:
        f.write('extrinsic\n')
        for i in range(0, 4):
            for j in range(0, 4):
                f.write(str(extrinsics[i][j]) + ' ')
            f.write('\n')
        f.write('\n')

        f.write('intrinsic\n')
        for i in range(0, 3):
            for j in range(0, 3):
                f.write(str(intrinsics[i][j]) + ' ')
            f.write('\n')

def write_matrix(M: np.ndarray, mat_file: str) -> None:
    """Writes a single matrix to a file.

    Parameters:
        M: Matrix to be stored.
        mat_file: Output file where the given matrix is to be writen.
    """
    with open(mat_file, "w") as f:
        for row in M:
            for e in row:
                f.write("{} ".format(e))
            f.write("\n")

def write_mesh(mesh_file: str, mesh: o3d.geometry.TriangleMesh) -> None:
    """Writes a mesh to a file.

    Parameters:
        mesh_file: Output mesh file.
        mesh: Mesh to be stored.
    """
    return o3d.io.write_triangle_mesh(mesh_file, mesh)

def write_pfm(pfm_file: str, data_map: np.ndarray, scale: float = 1.0) -> None:
    """Writes a data map to a file in *.pfm format.

    Parameters:
        pfm_file: Output *.pfm file to store the data map.
        data_map: Data map to be stored.
        scale: Value used to scale the data map.
    """
    with open(pfm_file, 'wb') as pfm_file:
        color = None

        if data_map.dtype.name != 'float32':
            raise Exception('Image dtype must be float32.')

        data_map = np.flipud(data_map)

        if len(data_map.shape) == 3 and data_map.shape[2] == 3: # color data_map
            color = True
        elif len(data_map.shape) == 2 or (len(data_map.shape) == 3 and data_map.shape[2] == 1): # greyscale
            color = False
        else:
            raise Exception('Image must have H x W x 3, H x W x 1 or H x W dimensions.')

        a = 'PF\n' if color else 'Pf\n'
        b = '%d %d\n' % (data_map.shape[1], data_map.shape[0])
        
        pfm_file.write(a.encode('iso8859-15'))
        pfm_file.write(b.encode('iso8859-15'))

        endian = data_map.dtype.byteorder

        if endian == '<' or endian == '=' and sys.byteorder == 'little':
            scale = -scale

        c = '%f\n' % scale
        pfm_file.write(c.encode('iso8859-15'))

        data_map_string = data_map.tostring()
        pfm_file.write(data_map_string)

def write_point_cloud(fn, cloud):
    o3d.io.write_point_cloud(fn, cloud)

def save_state_dict(model, save_path):
    torch.save(model.state_dict(), save_path)

def load(model, load_path):
    model.load_state_dict(torch.load(load_path))

def save_ckpt(model, save_path):
    save_dict = {"model": model.state_dict()}
    torch.save(save_dict, save_path)

def load_ckpt(model, load_path, strict=True):
    model_dict = torch.load(load_path)
    model.load_state_dict(model_dict['model'], strict=strict)

def save_model(model, cfg, name="ckpt_model.pth"):
    """Saves model weights to disk.
    """
    ckpt_path = cfg["model"]["ckpt"]
    if not os.path.exists(ckpt_path):
        os.makedirs(ckpt_path)
    print(f"Saving model checkpoint to {ckpt_path}...")
    model_path = os.path.join(save_folder, name)
    torch.save(model.state_dict(), model_path)

def load_pretrained_model(model, ckpt):
    """Loads model weights from disk.
    """
    print(f"Loading model from: {ckpt}...")
    try:
        model.load_state_dict(torch.load(ckpt))
    except Exception as e:
        print(e)
        print("Failed loading network weights...")
        sys.exit()



# cvt/visualization/util.py
"""Module including general utilities for visualization.

This module includes the following functions:


"""
import open3d as o3d
import numpy as np
import cv2
import scipy.ndimage as ndimage
import skimage.transform as transform
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.collections as mc
import torch
import torch.nn.functional as F
import os,sys
import torchvision.transforms.functional as tvf
import torchvision.transforms as tvt

from common import non_zero_std
from io import *

def display_inlier_outlier(cloud: o3d.geometry.PointCloud, indices: np.ndarray) -> None:
    """Displays a point cloud with outlier points colored red.

    Parameters:
        cloud: Point cloud to be displayed.
        indices: Indices indicating the inlier points.
    """
    inlier_cloud = cloud.select_by_index(ind)
    outlier_cloud = cloud.select_by_index(ind, invert=True)

    outlier_cloud.paint_uniform_color([1, 0, 0])
    inlier_cloud.paint_uniform_color([0.8, 0.8, 0.8])
    o3d.visualization.draw_geometries([inlier_cloud, outlier_cloud],
                                      zoom=0.3412,
                                      front=[0.4257, -0.2125, -0.8795],
                                      lookat=[2.6172, 2.0475, 1.532],
                                      up=[-0.0694, -0.9768, 0.2024])

    return

def print_csv(data):
    for i,d in enumerate(data):
        if i==len(data)-1:
            print(f"{d:6.4f}")
        else:
            print(f"{d:6.4f}", end=",")

def plot_cameras(cams, num_cams, scale, A, output_file):
    # grab the requested number of cameras and apply the alignment
    P = cams[:num_cams]
    P = np.array([ A @ np.linalg.inv(p) for p in P ])

    # create 'num_cams' intrinsic matrices (just used for point-cloud camera pyramid geometry)
    k = np.array([[233.202, 0.0, 144.753],[0.0, 233.202, 108.323],[0.0, 0.0, 1.0]])
    K = np.array([ k for p in P])

    # build list of camera pyramid points
    pyr_pts = []
    for k,p in zip(K,P):
        pyr_pt = build_cam_pyr(scale, k)
        pyr_pt = p @ pyr_pt
        pyr_pts.append(pyr_pt)

    # build point cloud using camera centers
    build_pyr_point_cloud(pyr_pts, output_file)

    return


f_matrix = None
img2_line = None
img1 = None
img2 = None

CENTERED = False


def mouse1_callback(event, x, y, flags, param):
    if event == cv2.EVENT_MOUSEMOVE:
        #print("mouse1 X: " + str(x) + " Y: " + str(y))
#        y = img1.shape[0] - y
        if CENTERED:
            x -= img1.shape[1] / 2
            y -= img1.shape[0] / 2
        mouse1_pt = np.asarray([x, y, 1.0])
        i2ray = np.dot(f_matrix, mouse1_pt)
        if CENTERED:
            i2ray[2] = i2ray[2] - i2ray[0]*img2.shape[1]/2 - i2ray[1]*img2.shape[0]/2 
        #print(i2ray)
        i2pt1_x = 0
        i2pt2_x = img2.shape[1]
        i2pt1_y = int(-(i2ray[2] + i2ray[0] * i2pt1_x) / i2ray[1])
        i2pt2_y = int(-(i2ray[2] + i2ray[0] * i2pt2_x) / i2ray[1])

        global img2_line
#        img2_line = ((i2pt1_x, img1.shape[0] - i2pt1_y), (i2pt2_x, img1.shape[0] - i2pt2_y))
        img2_line = ((i2pt1_x, i2pt1_y), (i2pt2_x, i2pt2_y))

def draw_line(img, line):
    if line is None:
        return img
    # Clip the line to image bounds
    ret, p1, p2 = cv2.clipLine((0, 0, img.shape[1], img.shape[0]), line[0], line[1])
    if ret:
        img = img.copy()
        cv2.line(img, p1, p2, (255, 255, 0), 1)
    return img


def scale_img(img, scale):
    img_scaled = transform.resize(img, [img.shape[0]*scale, img.shape[1]*scale], mode="constant")
    return img_scaled.astype(np.float32)

def scale_f_mat(mat, scale):
    mat[:, 2] *= scale
    mat[2, :] *= scale
    return mat

def fmat_demo(img1l, img2l, fl, scale=1.0):
    global img1, img2, f_matrix

    img1 = scale_img(img1l, scale)
    img2 = scale_img(img2l, scale)
    f_matrix = scale_f_mat(fl, scale)

    f_matrix = f_matrix/np.max(f_matrix)

    print("Img1: " + str(img1.shape))
    print("Img2: " + str(img2.shape))

    cv2.namedWindow("img1")
    cv2.namedWindow("img2")
    cv2.setMouseCallback("img1", mouse1_callback)

    while True:
        show2 = draw_line(img2, img2_line)
        show1 = img1
        cv2.imshow("img1", cv2.cvtColor(show1, cv2.COLOR_BGR2RGB))
        cv2.imshow("img2", cv2.cvtColor(show2, cv2.COLOR_BGR2RGB))
        cv2.waitKey(50)

def build_pyr_point_cloud(pyr_pts, filename):
    """Builds a point cloud for a camera frustum visual.

    !!! bug "needs work..."
    """
    num_pts = len(pyr_pts)
    element_vertex = 6*num_pts
    element_edge = 10*num_pts

    with open(filename, 'w') as fh:
        # write header meta-data
        fh.write('ply\n')
        fh.write('format ascii 1.0\n')
        fh.write('comment Right-Handed System\n')
        fh.write('element vertex {}\n'.format(element_vertex))
        fh.write('property float x\n')
        fh.write('property float y\n')
        fh.write('property float z\n')
        fh.write('property uchar red\n')
        fh.write('property uchar green\n')
        fh.write('property uchar blue\n')
        fh.write('element edge {}\n'.format(element_edge))
        fh.write('property int vertex1\n')
        fh.write('property int vertex2\n')
        fh.write('property uchar red\n')
        fh.write('property uchar green\n')
        fh.write('property uchar blue\n')
        fh.write('end_header\n')

        # write vertex data to file
        for pt in pyr_pts:
            fh.write('{:.10f}'.format( pt[0][0][0] ) + ' ' + '{:.10f}'.format( pt[0][1][0] ) + ' ' + '{:.10f}'.format( pt[0][2][0] ) + ' 255 128 0\n')
            fh.write('{:.10f}'.format( pt[1][0][0] ) + ' ' + '{:.10f}'.format( pt[1][1][0] ) + ' ' + '{:.10f}'.format( pt[1][2][0] ) + ' 255 128 0\n')
            fh.write('{:.10f}'.format( pt[2][0][0] ) + ' ' + '{:.10f}'.format( pt[2][1][0] ) + ' ' + '{:.10f}'.format( pt[2][2][0] ) + ' 255 128 0\n')
            fh.write('{:.10f}'.format( pt[3][0][0] ) + ' ' + '{:.10f}'.format( pt[3][1][0] ) + ' ' + '{:.10f}'.format( pt[3][2][0] ) + ' 255 128 0\n')
            fh.write('{:.10f}'.format( pt[4][0][0] ) + ' ' + '{:.10f}'.format( pt[4][1][0] ) + ' ' + '{:.10f}'.format( pt[4][2][0] ) + ' 255 128 0\n')
            fh.write('{:.10f}'.format( pt[5][0][0] ) + ' ' + '{:.10f}'.format( pt[5][1][0] ) + ' ' + '{:.10f}'.format( pt[5][2][0] ) + ' 255 128 0\n')

        # write edge data to file
        for i in range(num_pts):
            edge_ind = i*6
            fh.write('{} {} 255 0 0\n'.format(edge_ind, edge_ind+1))
            fh.write('{} {} 255 0 0\n'.format(edge_ind, edge_ind+2))
            fh.write('{} {} 255 0 0\n'.format(edge_ind, edge_ind+3))
            fh.write('{} {} 255 0 0\n'.format(edge_ind, edge_ind+4))
            fh.write('{} {} 255 0 0\n'.format(edge_ind+1, edge_ind+2))
            fh.write('{} {} 255 0 0\n'.format(edge_ind+2, edge_ind+3))
            fh.write('{} {} 255 0 0\n'.format(edge_ind+3, edge_ind+4))
            fh.write('{} {} 255 0 0\n'.format(edge_ind+4, edge_ind+1))
            fh.write('{} {} 255 0 0\n'.format(edge_ind+1, edge_ind+5))
            fh.write('{} {} 255 0 0\n'.format(edge_ind+5, edge_ind+2))
    return

def build_cam_pyr(cam_scale, K):
    """Constructs a camera frustum for visualization.

    !!! bug "needs work..."
    """
    focallen   = K[0][0]
    cam_w      = 2 * K[0][2]
    cam_h      = 2 * K[1][2]
    cam_center = np.array([0.0,          0.0,          0.0,      1.0])
    cam_ul     = np.array([cam_w * -0.5, cam_h * -0.5, focallen, 1.0])
    cam_ur     = np.array([cam_w *  0.5, cam_h * -0.5, focallen, 1.0])
    cam_dr     = np.array([cam_w *  0.5, cam_h *  0.5, focallen, 1.0])
    cam_dl     = np.array([cam_w * -0.5, cam_h *  0.5, focallen, 1.0])
    cam_top    = np.array([0.0,          cam_h * -0.7, focallen, 1.0])
    cam_center *= cam_scale
    cam_ul     *= cam_scale
    cam_ur     *= cam_scale
    cam_dr     *= cam_scale
    cam_dl     *= cam_scale
    cam_top    *= cam_scale
    cam_center[3] = 1.0
    cam_ul[3]     = 1.0
    cam_ur[3]     = 1.0
    cam_dr[3]     = 1.0
    cam_dl[3]     = 1.0
    cam_top[3]    = 1.0
    cam_center = cam_center.reshape((4, 1))
    cam_ul     = cam_ul.reshape((4, 1))
    cam_ur     = cam_ur.reshape((4, 1))
    cam_dr     = cam_dr.reshape((4, 1))
    cam_dl     = cam_dl.reshape((4, 1))
    cam_top    = cam_top.reshape((4, 1))
    return [cam_center, cam_ul, cam_ur, cam_dr, cam_dl, cam_top]


def display_map(filename: str, disp_map: np.ndarray, mx: float, mn: float) -> None:
    """Writes an input map to a normalized image.

    Parameters:
        filename: Name of the file to store the input map.
        disp_map: Map to be stored as an image file.
        mx: maximum value used for pixel intensity normalization.
        mn: minimum value used for pixel intensity normalization.
    """
    disp_map = ((disp_map-mn)/(mx-mn+1e-8))*255
    cv2.imwrite(filename, disp_map)

def plot_coverage(data, output, batch_ind, vis_path):
    hypos = output["hypos"]
    intervals = output["intervals"]
    _,H,W = data["target_depth"].shape
    levels = len(hypos)

    uncovered_mask
    for l in range(levels):
        hypo = hypos[l].squeeze(1)
        batch_size, planes, h, w = hypo.shape
        interval = intervals[l].squeeze(1)
        target_depth = tvf.resize(data["target_depth"], [h,w]).unsqueeze(1)


        ### compute coverage
        diff = torch.abs(hypo - target_depth)
        min_interval = interval[:,0:1] * 0.5 # intervals are bin widths, divide by 2 for radius
        coverage = torch.clip(torch.where(diff <= min_interval, 1, 0).sum(dim=1, keepdim=True), 0, 1)
        uncovered = torch.clip(torch.where(coverage <= 0, 1, 0).sum(dim=1, keepdim=True), 0, 1)
        valid_targets = torch.where(target_depth > 0, 1, 0)
        uncovered *= valid_targets
        cov_percent = coverage.sum() / (valid_targets.sum() + 1e-10)
        coverage = coverage.reshape(h,w,1).repeat(1,1,3) * (torch.tensor([[[0,255,0]]]).to(coverage).repeat(h,w,1))
        uncovered = uncovered.reshape(h,w,1).repeat(1,1,3) * (torch.tensor([[[255,0,0]]]).to(uncovered).repeat(h,w,1))

        # plot
        total_coverage = torch.movedim(coverage+uncovered, (0,1,2), (1,2,0))
        total_coverage = torch.movedim(tvf.resize(total_coverage, [H,W], interpolation=tvt.InterpolationMode.NEAREST), (0,1,2), (2,0,1))
        plt.imshow((total_coverage).detach().cpu().numpy())
        plt.title(f"Coverage ({cov_percent*100:0.2f}%)")
        plt.axis('off')
        plt.savefig(os.path.join(vis_path, f"coverage_{batch_ind:08d}_l{l}.png"))
        plt.close()

def visualize_relative_variance(target_cost, cost_volume, hypos, intervals, target_depth, mask, batch_ind, vis_path, level):
    batch_size, _, planes, h, w = cost_volume.shape
    batch_ind = int(batch_ind.item())
    cv = cost_volume.squeeze(dim=1)
    mask = mask.reshape(1, h, w).repeat(batch_size, 1, 1)
    hypos = hypos.reshape(batch_size, planes, h, w)
    intervals = intervals.reshape(batch_size, planes, h, w)
    target_depth = target_depth.reshape(batch_size, 1, h, w)

    ### compute coverage
    diff = torch.abs(hypos - target_depth)
    min_interval = intervals[:,0:1] * 0.5 # intervals are bin widths, divide by 2 for radius
    coverage = torch.clip(torch.where(diff <= min_interval, 1, 0).sum(dim=1, keepdim=True), 0, 1)
    uncovered = torch.clip(torch.where(coverage <= 0, 1, 0).sum(dim=1, keepdim=True), 0, 1)
    valid_targets = torch.where(target_depth > 0, 1, 0)
    uncovered *= valid_targets
    cov_percent = coverage.sum() / (valid_targets.sum() + 1e-10)
    coverage = coverage.reshape(h,w,1).repeat(1,1,3) * (torch.tensor([[[0,255,0]]]).to(coverage).repeat(h,w,1))
    uncovered = uncovered.reshape(h,w,1).repeat(1,1,3) * (torch.tensor([[[255,0,0]]]).to(uncovered).repeat(h,w,1))
    plt.imshow((coverage+uncovered).detach().cpu().numpy())
    plt.title(f"Coverage ({cov_percent*100:0.2f}%)")
    plt.axis('off')
    plt.savefig(os.path.join(vis_path,f"coverage_{batch_ind:08d}_l{level}.png"))
    plt.close()

    min_interval = intervals[:,0:1].repeat(1, planes, 1, 1)
    tcv = target_cost.reshape(batch_size, 1, h, w).repeat(1, planes, 1, 1)
    mask = mask.reshape(batch_size, 1, h, w).repeat(1, planes, 1, 1)
    gt_bin_mask = torch.where(diff <= min_interval, 0, 1)
    mask *= gt_bin_mask
    inds = torch.stack(torch.where(mask > 0), dim=-1)
    var_diff = torch.div(tcv,cv+1e-10)[inds[:,0],inds[:,1],inds[:,2],inds[:,3]]

    # compute mean and standard-deviation
    sigma = torch.std(var_diff)
    mean = torch.mean(var_diff)
    # select only values in 3-sigma radius
    var_diff = var_diff[torch.where(var_diff >= (mean-(3*sigma)))[0]]
    var_diff = var_diff[torch.where(var_diff <= (mean+(3*sigma)))[0]]
    # plot
    flat_var = var_diff.detach().cpu().numpy().flatten()
    plt.hist(flat_var, bins=100)
    plt.axvline(flat_var.mean(), c="red")
    plt.savefig(os.path.join(vis_path,f"relative_{batch_ind:08d}_l{level}.png"))
    plt.close()

    return mean, cov_percent

def visualize_point_variance(target_cost, cost_volume, hypos, intervals, target_depth, mask, batch_ind, vis_path, level):
    batch_size, groups, planes, h, w = cost_volume.shape
    batch_ind = int(batch_ind.item())
    tcv = target_cost.mean(dim=1, keepdim=True)
    cv = cost_volume.mean(dim=1, keepdim=True)
    mask = mask.reshape(1, h, w).repeat(batch_size, 1, 1)
    hypos = hypos.reshape(batch_size, planes, h, w)
    intervals = intervals.reshape(batch_size, planes, h, w)
    target_depth = target_depth.reshape(batch_size, 1, h, w)

    ### compute coverage
    diff = torch.abs(hypos - target_depth)
    min_interval = intervals[:,0:1]
    coverage = torch.clip(torch.where(diff <= min_interval, 1, 0).sum(dim=1, keepdim=True), 0, 1)
    uncovered = torch.clip(torch.where(coverage <= 0, 1, 0).sum(dim=1, keepdim=True), 0, 1)
    valid_targets = torch.where(target_depth > 0, 1, 0)
    uncovered *= valid_targets
    cov_percent = coverage.sum() / (valid_targets.sum() + 1e-10)
    coverage = coverage.reshape(h,w,1).repeat(1,1,3) * (torch.tensor([[[0,255,0]]]).to(coverage).repeat(h,w,1))
    uncovered = uncovered.reshape(h,w,1).repeat(1,1,3) * (torch.tensor([[[255,0,0]]]).to(uncovered).repeat(h,w,1))
    plt.imshow((coverage+uncovered).detach().cpu().numpy())
    plt.title(f"Coverage ({cov_percent*100:0.2f}%)")
    plt.axis('off')
    plt.savefig(os.path.join(vis_path,f"coverage_{batch_ind:08d}_l{level}.png"))
    plt.close()

    ### compute correct matching variance histogram
    var_match = tcv.reshape(batch_size, h, w)
    inds = torch.stack(torch.where(mask > 0), dim=-1)
    var_match = var_match[inds[:,0], inds[:,1], inds[:,2]]
    # compute mean and standard-deviation
    match_sigma = torch.std(var_match)
    match_mean = torch.mean(var_match)
    # select only values in 3-sigma radius
    var_match = var_match[torch.where(var_match >= (match_mean-(3*match_sigma)))[0]]
    var_match = var_match[torch.where(var_match <= (match_mean+(3*match_sigma)))[0]]
    # plot
    flat_var = var_match.detach().cpu().numpy().flatten()
    plt.hist(flat_var, bins=100)
    plt.axvline(flat_var.mean(), c="red")
    plt.savefig(os.path.join(vis_path,f"match_{batch_ind:08d}_l{level}.png"))
    plt.close()

    ### compute incorrect matching variance histogram
    mask = mask.reshape(batch_size, 1, h, w).repeat(1, planes, 1, 1)
    inds = torch.stack(torch.where(mask > 0), dim=-1)
    diff = diff[inds[:,0], inds[:,1], inds[:,2], inds[:,3]]
    min_interval = min_interval[inds[:,0], 0, inds[:,2], inds[:,3]]
    cv = cv[inds[:,0], 0, inds[:,1], inds[:,2], inds[:,3]]
    inds = torch.stack(torch.where(diff > min_interval), dim=-1)
    var_mismatch = cv[inds]
    # compute mean and standard-deviation
    mismatch_sigma = torch.std(var_mismatch)
    mismatch_mean = torch.mean(var_mismatch)
    # select only values in 3-sigma radius
    var_mismatch = var_mismatch[torch.where(var_mismatch >= (mismatch_mean-(3*mismatch_sigma)))[0]]
    var_mismatch = var_mismatch[torch.where(var_mismatch <= (mismatch_mean+(3*mismatch_sigma)))[0]]
    # plot
    flat_var = var_mismatch.detach().cpu().numpy().flatten()
    plt.hist(flat_var, bins=100)
    plt.axvline(flat_var.mean(), c="red")
    plt.savefig(os.path.join(vis_path,f"mismatch_{batch_ind:08d}_l{level}.png"))
    plt.close()

    return match_mean, mismatch_mean

def visualize_ray_points(rays, ind, edge_color="255 0 0"):
    out_file = f"cam_vis/ray_points_{ind:04d}.ply"
    num_rays, num_points, _ = rays.shape

    with open(out_file, "w") as of:
        of.write("ply\n")
        of.write("format ascii 1.0\n")
        of.write("comment VCGLIB generated\n")
        of.write(f"element vertex {num_rays*num_points}\n")
        of.write("property float x\n")
        of.write("property float y\n")
        of.write("property float z\n")
        of.write("property uchar red\n")
        of.write("property uchar green\n")
        of.write("property uchar blue\n")
        of.write("end_header\n")
        for i in range(num_rays):
            for j in range(num_points):
                of.write(f"{rays[i,j,0]:0.3f} {rays[i,j,1]:0.3f} {rays[i,j,2]:0.3f} 0 0 0\n")

def visualize_camera_frustum(planes, ind, edge_color="255 0 0"):
    out_file = f"cam_vis/cam_frustum_{ind:04d}.ply"
    num_planes, _, _ = planes.shape
    num_verts = 4*num_planes
    num_edges = 8*num_planes - 4

    with open(out_file, "w") as of:
        of.write("ply\n")
        of.write("format ascii 1.0\n")
        of.write("comment VCGLIB generated\n")
        of.write(f"element vertex {num_verts}\n")
        of.write("property float x\n")
        of.write("property float y\n")
        of.write("property float z\n")
        of.write("property uchar red\n")
        of.write("property uchar green\n")
        of.write("property uchar blue\n")
        of.write(f"element edge {num_edges}\n")
        of.write("property int vertex1\n")
        of.write("property int vertex2\n")
        of.write("property uchar red\n")
        of.write("property uchar green\n")
        of.write("property uchar blue\n")
        of.write("end_header\n")
        for p in range(num_planes):
            for i in range(4):
                of.write(f"{planes[p,0,i]:0.3f} {planes[p,1,i]:0.3f} {planes[p,2,i]:0.3f} 0 0 0\n")

        # draw plane border
        for p in range(num_planes):
            ind = p*4
            of.write(f"{ind} {ind+1} {edge_color}\n")
            of.write(f"{ind} {ind+2} {edge_color}\n")
            of.write(f"{ind+1} {ind+3} {edge_color}\n")
            of.write(f"{ind+2} {ind+3} {edge_color}\n")

        # draw plane connections
        for p in range(num_planes-1):
            ind = p*4
            of.write(f"{ind} {ind+4} {edge_color}\n")
            of.write(f"{ind+1} {ind+5} {edge_color}\n")
            of.write(f"{ind+2} {ind+6} {edge_color}\n")
            of.write(f"{ind+3} {ind+7} {edge_color}\n")

def visualize_bounding_box(min_bounds, max_bounds, output_file, edge_color="255 0 0"):

    num_verts = 8
    num_edges = 12
    with open(out_file, "w") as of:
        of.write("ply\n")
        of.write("format ascii 1.0\n")
        of.write("comment VCGLIB generated\n")
        of.write(f"element vertex {num_verts}\n")
        of.write("property float x\n")
        of.write("property float y\n")
        of.write("property float z\n")
        of.write("property uchar red\n")
        of.write("property uchar green\n")
        of.write("property uchar blue\n")
        of.write(f"element edge {num_edges}\n")
        of.write("property int vertex1\n")
        of.write("property int vertex2\n")
        of.write("property uchar red\n")
        of.write("property uchar green\n")
        of.write("property uchar blue\n")
        of.write("end_header\n")
        for p in range(num_planes):
            for i in range(4):
                of.write(f"{planes[p,0,i]:0.3f} {planes[p,1,i]:0.3f} {planes[p,2,i]:0.3f} 0 0 0\n")

        # draw plane border
        for p in range(num_planes):
            ind = p*4
            of.write(f"{ind} {ind+1} {edge_color}\n")
            of.write(f"{ind} {ind+2} {edge_color}\n")
            of.write(f"{ind+1} {ind+3} {edge_color}\n")
            of.write(f"{ind+2} {ind+3} {edge_color}\n")

        # draw plane connections
        for p in range(num_planes-1):
            ind = p*4
            of.write(f"{ind} {ind+4} {edge_color}\n")
            of.write(f"{ind+1} {ind+5} {edge_color}\n")
            of.write(f"{ind+2} {ind+6} {edge_color}\n")
            of.write(f"{ind+3} {ind+7} {edge_color}\n")



def visualize_mvs(data, output, batch_ind, vis_path, max_depth_error, mode, epoch=-1):
    image = torch.movedim(data["images"][0,0], (0,1,2), (2,0,1)).detach().cpu().numpy()
    image = ((image-image.min()) / (image.max()-image.min()+1e-10))

    target_depth = data["target_depth"].detach().cpu().numpy()[0,0]
    est_depth = output["final_depth"].detach().cpu().numpy()[0,0]
    est_conf = output["confidence"].detach().cpu().numpy()[0,0]

    assert(est_depth.shape == target_depth.shape)

    num_valid_pixels = np.where(target_depth > 0, 1, 0).sum()

    ## compute depth residual
    depth_residual = np.abs(est_depth - target_depth)
    res_temp = np.clip(depth_residual / (max_depth_error*3+1e-10), 0, 1)
    depth_residual[target_depth == 0.0] = 0.0
    depth_mae = np.mean(depth_residual)

    ## compute ROC and AUC
    perc, oracle_roc, est_roc, rel_auc = auc_score(data, output)

    ### plot coverage
    #levels = uncovered_masks.shape[0]
    #fig, axs = plt.subplots(1, levels)
    #fig.tight_layout()
    #for i in range(levels):
    #    coverage_percent = 1 - (uncovered_masks[i].sum() / num_valid_pixels)
    #    axs[i].imshow(uncovered_masks[i], cmap="gray", vmin=0.0, vmax=1.0)
    #    axs[i].set_title(f"Level {i}: \n{coverage_percent*100:0.2f}%")
    #    axs[i].set_xticks([])
    #    axs[i].set_yticks([])
    #plt.subplots_adjust(wspace=0, hspace=0.2)
    #plot_file = os.path.join(vis_path, f"{batch_ind:08d}_cov.png")
    #plt.savefig(plot_file, bbox_inches='tight', pad_inches=0.4, dpi=400)
    #plt.clf()
    #plt.close()

    ## plot
    fig, axs = plt.subplots(2, 3)
    fig.tight_layout()
    max_depth = np.max(target_depth)
    # Row #1: Depth
    axs[0, 0].imshow(target_depth, cmap="gray", vmin=0, vmax=max_depth)
    axs[0, 0].set_title('Target Depth')
    axs[0, 0].set_xticks([])
    axs[0, 0].set_yticks([])
    axs[0, 1].imshow(est_depth, cmap="gray", vmin=0, vmax=max_depth)
    axs[0, 1].set_title('Est. Depth')
    axs[0, 1].set_xticks([])
    axs[0, 1].set_yticks([])
    axs[0, 2].imshow(depth_residual, cmap="hot", vmin=0, vmax=max_depth_error)
    axs[0, 2].set_title(f'Residual (mae: {depth_mae:0.3f}mm)')
    axs[0, 2].set_xticks([])
    axs[0, 2].set_yticks([])
    # Row #2: Confidence
    axs[1, 0].imshow(image)
    axs[1, 0].set_title('Image')
    axs[1, 0].set_xticks([])
    axs[1, 0].set_yticks([])
    axs[1, 1].imshow(est_conf, cmap="gray", vmin=0, vmax=1.0)
    axs[1, 1].set_title('Est. Confidence')
    axs[1, 1].set_xticks([])
    axs[1, 1].set_yticks([])
    axs[1, 2].plot(perc, oracle_roc, label="Oracle")
    axs[1, 2].plot(perc, est_roc, label="Confidence")
    axs[1, 2].set_title(f'ROC (rel. auc: {rel_auc:0.2f})')
    axs[1, 2].set_xticks([])
    axs[1, 2].set_yticks([])
    axs[1, 2].legend()
    plt.subplots_adjust(wspace=0, hspace=0.2)

    if epoch!=None and epoch >= 0:
        plot_file = os.path.join(vis_path, f"{mode}_{epoch:03d}_{batch_ind:08d}.png")
    else:
        plot_file = os.path.join(vis_path, f"{batch_ind:08d}.png")
    plt.savefig(plot_file, bbox_inches='tight', pad_inches=0.4, dpi=400)
    plt.clf()
    plt.close()

def auc_score(data, output):
    est_depth = output["final_depth"][0,0]
    est_conf = output["confidence"][0,0]
    target_depth = data["target_depth"][0,0]

    mask = torch.where(target_depth > 0, 1.0, 0.0)
    pixel_count = int(torch.sum(mask).item())

    inds = torch.where(mask==1)
    target_depth = target_depth[inds].detach().cpu().numpy()
    est_depth = est_depth[inds].detach().cpu().numpy()
    est_conf = est_conf[inds].detach().cpu().numpy()

    # flatten to 1D tensor
    target_depth = target_depth.flatten()
    est_depth = est_depth.flatten()
    est_conf = est_conf.flatten()

    # compute error
    error = np.abs(est_depth - target_depth)
    
    # sort orcale curves by error
    oracle_indices = np.argsort(error)
    oracle_error = np.take(error, indices=oracle_indices, axis=0)

    # sort all tensors by confidence value
    est_indices_conf = np.argsort(est_conf)
    est_indices_conf = est_indices_conf[::-1]
    est_error_conf = np.take(error, indices=est_indices_conf, axis=0)

    # build density vector
    perc = np.array(list(range(5,105,5)))
    density = np.array((perc/100) * (pixel_count), dtype=np.int32)

    oracle_roc = np.zeros(density.shape)
    est_roc = np.zeros(density.shape)
    for i,k in enumerate(density):
        oe = oracle_error[:k]
        ee = est_error_conf[:k]

        if (oe.shape[0] == 0):
            oracle_roc[i] = 0.0
            est_roc[i] = 0.0
        else:
            oracle_roc[i] = np.mean(oe)
            est_roc[i] = np.mean(ee)

    # comput AUC
    oracle_auc = np.trapz(oracle_roc, dx=1)
    est_auc = np.trapz(est_roc, dx=1)

    return perc, oracle_roc, est_roc, (est_auc/oracle_auc)


def laplacian_count(data, output, plot_file=None, use_est_depth=False):
    target_depth = data["target_depth"].detach().cpu().numpy()[0]
    image_laplacian = output["image_laplacian"][0].detach().cpu().numpy()
    if use_est_depth:
        depth_laplacian = output["est_depth_laplacian"][0].detach().cpu().numpy()
    else:
        depth_laplacian = data["depth_laplacian"][0].detach().cpu().numpy()

    # Image Laplacian vs. Depth Laplacian Count
    inds = np.argwhere(target_depth.flatten() > 0.0)
    num_pix = int(inds.shape[0])
    il = (image_laplacian.flatten())[inds][:,0]
    dl = (depth_laplacian.flatten())[inds][:,0]
    M = np.zeros((5,5))
    for i in range(5):
        for d in range(5):
            M[i,d] = (int(((il == i) & (dl == d)).sum()) / num_pix)

    if plot_file != None:
        fig,ax = plt.subplots()
        img = ax.imshow(M, interpolation="none", cmap="copper")
        ax.set_xlabel("Est. Depth Laplacian")
        ax.set_ylabel("Image Laplacian")
        ax.set_xticks(np.arange(5))
        ax.set_yticks(np.arange(5))
        for i in range(5):
            for j in range(5):
                text = ax.text(i, j, f"{(M[j, i]*100):0.2f}%", ha="center", va="center", color="w")
        plt.savefig(plot_file, bbox_inches='tight', pad_inches=0.4, dpi=200)
        plt.clf()
        plt.close()

    return M


def laplacian_uncovered_count(data, output, plot_file=None):
    image_laplacian = output["image_laplacian"][0].detach().cpu().numpy()
    depth_laplacian = data["depth_laplacian"][0].detach().cpu().numpy()
    uncovered_masks = output["uncovered_masks"]

    # Image Laplacian vs. Depth Laplacian vs. Uncovered Count
    for mask in uncovered_masks:
        inds = np.argwhere(mask.flatten() > 0.0)
        num_pix = int(inds.shape[0])
        il = (image_laplacian.flatten())[inds][:,0]
        dl = (depth_laplacian.flatten())[inds][:,0]
        M = np.zeros((5,5))
        for i in range(5):
            for d in range(5):
                M[i,d] += (int(((il == i) & (dl == d)).sum()) / num_pix)

    if plot_file != None:
        fig,ax = plt.subplots()
        img = ax.imshow(M, interpolation="none", cmap="copper")
        ax.set_xlabel("Depth Laplacian")
        ax.set_ylabel("Image Laplacian")
        ax.set_xticks(np.arange(5))
        ax.set_yticks(np.arange(5))
        for i in range(5):
            for j in range(5):
                text = ax.text(i, j, f"{M[j, i]:0.2f}", ha="center", va="center", color="w")
        plt.savefig(plot_file, bbox_inches='tight', pad_inches=0.4, dpi=200)
        plt.clf()
        plt.close()

    return M

def laplacian_depth_error(data, output, plot_file=None, use_est_depth=False):
    target_depth = data["target_depth"].detach().cpu().numpy()[0,0]
    est_depth = output["final_depth"].detach().cpu().numpy()[0,0]
    image_laplacian = output["image_laplacian"][0].detach().cpu().numpy()
    if use_est_depth:
        depth_laplacian = output["est_depth_laplacian"][0].detach().cpu().numpy()
    else:
        depth_laplacian = data["depth_laplacian"][0].detach().cpu().numpy()

    abs_err = np.abs((est_depth - target_depth))
    inds = np.argwhere(target_depth.flatten() > 0.0)
    err = (abs_err.flatten())[inds]
    il = (image_laplacian.flatten())[inds][:,0]
    dl = (depth_laplacian.flatten())[inds][:,0]

    M = np.zeros((5,5))
    for i in range(5):
        for d in range(5):
            M[i,d] = err[np.argwhere((il == i) & (dl == d))].mean()

    if plot_file != None:
        fig,ax = plt.subplots()
        img = ax.imshow(M, interpolation="none", cmap="copper")
        ax.set_xlabel("Depth Laplacian")
        ax.set_ylabel("Image Laplacian")
        ax.set_xticks(np.arange(5))
        ax.set_yticks(np.arange(5))
        for i in range(5):
            for j in range(5):
                text = ax.text(i, j, f"{M[j, i]:0.2f}", ha="center", va="center", color="w")
        plt.savefig(plot_file, bbox_inches='tight', pad_inches=0.4, dpi=200)
        plt.clf()
        plt.close()

    return M

def plot_laplacian_matrix(M, plot_file, use_est_depth=False, count=False):
    fig,ax = plt.subplots()
    img = ax.imshow(M, interpolation="none", cmap="copper")
    if use_est_depth:
     ax.set_xlabel("Est. Depth Laplacian")
    else:
     ax.set_xlabel("Depth Laplacian")
    ax.set_ylabel("Image Laplacian")
    ax.set_xticks(np.arange(5))
    ax.set_yticks(np.arange(5))
    for i in range(5):
     for j in range(5):
         if count:
             text = ax.text(i, j, f"{(M[j, i]*100):0.2f}%", ha="center", va="center", color="w")
         else:
             text = ax.text(i, j, f"{M[j, i]:0.2f}", ha="center", va="center", color="w")
    plt.savefig(plot_file, bbox_inches='tight', pad_inches=0.4, dpi=200)
    plt.clf()
    plt.close()


def to_normal(ply_file, output_file, radius=15.0, max_nn=100):
    cloud = o3d.io.read_point_cloud(ply_file)
    cloud.estimate_normals(search_param=o3d.geometry.KDTreeSearchParamHybrid(radius=radius, max_nn=max_nn))
    normals = np.asarray(cloud.normals)
    #normals = (normals-normals.min(axis=0)) / (normals.max(axis=0)-normals.min(axis=0))
    normals = (normals-normals.min()) / (normals.max()-normals.min())
    cloud.colors = o3d.utility.Vector3dVector(normals[:,::-1])
    o3d.io.write_point_cloud(output_file, cloud)



# cvt/util.py

"""A suite of common utility functions.

This module includes general utility functions used
by the sub-packages of the CVTkit library.

This module contains the following functions:

- 'build_coords_list(H, W, batch_size, device)' - Constructs an batched index list of pixel coordinates.
- 'compute_laplacian_pyr(image, levels=4)' - Computes the Laplacian pyramid of an image.
- 'cosine_similarity(v1, v2)' - Computes the cosine similarity between two vectors.
- 'groupwise_correlation(v1, v2, num_groups)' - 
- `non_zero_std(maps, device, dim, keepdim)` - Computes the standard deviation of all non-zero values in an input Tensor along the given dimension.
- 'parameters_count(net, name, do_print=True)' - 
- `print_gpu_mem()` - Prints the current unallocated memory of the GPU.
- `round_nearest(num, decimal=0)` - Rounds a floating point number to the nearest decimal place.
- 'scale_camera(cam, scale)' - 
- `scale_image(image, scale, interpolation)` - Scales an input pixel grid.
- 'scale_mvs_data(depths, confs, cams, scale, interpolation)' - 
- 'set_random_seed(seed)' - 
- 'to_gpu(data, device)' - 
- 'top_k_hypothesis_selection(grid, k, prev_hypos, prev_hypo_coords, prev_intervals)' - 
"""

import torch
import numpy as np
import cv2
import random
from typing import Tuple
import torch.nn.functional as F

# custom libraries
from camera import Z_from_disp, intrinsic_pyramid


def build_coords_list(H: int, W: int, batch_size: int, device: str) -> torch.Tensor:
    """Constructs an batched index list of pixel coordinates.

    Parameters:
        H: Height of the pixel grid.
        W: Width of the pixel grid.
        batch_size: Number of batches.
        device: GPU device identifier.

    Returns:
        The index list of shape [batch_size, H*W, 2]
    """
    indices_h = torch.linspace(0, H, H, dtype=torch.int64)
    indices_w = torch.linspace(0, W, W, dtype=torch.int64)
    indices_h, indices_w = torch.meshgrid(indices_h, indices_w)
    indices = torch.stack([indices_h, indices_w], dim=-1).to(torch.int64).to(device)
    indices = indices.reshape(1,-1,2).repeat(batch_size,1,1)
    return indices

def _build_depth_pyramid(depth, levels):
    h,w = depth.shape
    
    depths = {(levels-1): depth.reshape(1,h,w)}
    for i in range(1,levels):
        size = (int(w//(2**i)), int(h//(2**i)))
        d = cv2.resize(depth, size, cv2.INTER_LINEAR)

        depths[levels-1-i] = d.reshape(1, size[1], size[0])

    return depths

def build_labels(depth, hypotheses):
    bin_radius = (hypotheses[:,1] - hypotheses[:,0])/2.0
    target_bin_dist = torch.abs(depth - hypotheses)
    target_labels = torch.where(target_bin_dist <= bin_radius.unsqueeze(1), 1.0, 0.0)
    mask = torch.where(target_labels.sum(dim=1) > 0, 1.0, 0.0) * torch.where(depth.squeeze(1) > 0, 1.0, 0.0)
    target_labels = torch.argmax(target_labels, dim=1)

    return target_labels.to(torch.int64), mask.to(torch.float32)

def crop_image(image, crop_row, crop_col, scale):
    _, height, width = image.shape

    start_row = crop_row
    end_row = int(crop_row + (scale*height))
    start_col = crop_col
    end_col = int(crop_col + (scale*width))

    return image[:, start_row:end_row, start_col:end_col]

def freeze_model_weights(model):
    model.requires_grad_(False)

def laplacian_pyramid_th(image: torch.Tensor, tau: float) -> torch.Tensor:
    """Computes the Laplacian pyramid of an image.

    Parameters:
        image: 2D map to compute Laplacian over.
        tau: Laplacian region threshold.

    Returns:
        The map of the Laplacian regions.
    """
    batch_size, c, h, w = image.shape
    levels = 4

    # build gaussian pyramid
    pyr = [image]
    for l in range(levels):
        pyr.append(F.interpolate(pyr[-1], scale_factor=0.5, mode="bilinear"))

    # compute laplacian pyramid (differance between gaussian pyramid levels)
    for l in range(levels, 0, -1):
        region_id = (levels-l+1)

        diff = (torch.abs(F.interpolate(pyr[l], scale_factor=2, mode="bilinear") - pyr[l-1])).mean(dim=1, keepdim=True)
        diff = F.interpolate(diff, size=(h, w), mode="bilinear")
        diff_mask = torch.where(diff > tau, 1, 0)

        diff_mask = diff_mask.reshape(batch_size,h,w,1)

        if l==levels:
            all_diff_mask = diff_mask*region_id
        else:
            all_diff_mask = torch.where(diff_mask==1, region_id, all_diff_mask)

    return all_diff_mask.reshape(batch_size, 1, h, w)

def laplacian_pyramid(image: torch.Tensor) -> torch.Tensor:
    """Computes the Laplacian pyramid of an image.

    Parameters:
        image: 2D map to compute Laplacian over.
        tau: Laplacian region threshold.

    Returns:
        The map of the Laplacian regions.
    """
    batch_size, c, h, w = image.shape
    levels = 4

    # build gaussian pyramid
    pyr = [image]
    for l in range(levels):
        pyr.append(F.interpolate(pyr[-1], scale_factor=0.5, mode="bilinear"))

    # compute laplacian pyramid (differance between gaussian pyramid levels)
    for l in range(levels, 0, -1):
        region_id = (levels-l+1)

        diff = (torch.abs(F.interpolate(pyr[l], scale_factor=2, mode="bilinear") - pyr[l-1])).mean(dim=1, keepdim=True)
        diff = F.interpolate(diff, size=(h, w), mode="bilinear")
        diff = diff.reshape(batch_size,h,w,1)

        if l==levels:
            all_diff = diff
        else:
            all_diff += diff

    return all_diff.reshape(batch_size, 1, h, w)

def cosine_similarity(t1: torch.Tensor, t2: torch.Tensor) -> torch.Tensor:
    """Computes the cosine similarity between two tensors.

    Parameters:
        t1: First tensor.
        t2: Second tensor.

    Returns:
       The cosine similarity between the two tensors.
    """
    assert(t1.shape==t2.shape)
    similarity = torch.abs(F.cosine_similarity(t1,t2,dim=1).unsqueeze(1))

    if (len(t1.shape) == 5):
        B, C, D, H, W = t1.shape
        assert(similarity.shape == (B, 1, D, H, W))
    elif (len(t1.shape) == 4):
        B, C, H, W = t1.shape
        assert(similarity.shape == (B, 1, H, W))
    else:
        print("Can only compute cosine similarity with 4 or 5 dimension tensors")
        sys.exit()

    return similarity


def groupwise_correlation(t1: torch.Tensor, t2: torch.Tensor, num_groups: int) -> torch.Tensor:
    """Computes the Group-Wise Correlation (GWC) between two tensors.

    Parameters:
        t1: First tensor.
        t2: Second tensor.
        num_groups: Number of groups.

    Returns:
       The Group-Wise Correlation (GWC) between the two tensors.
    """
    assert(t1.shape==t2.shape)
    if (len(t1.shape) == 5):
        B, C, D, H, W = t1.shape
        assert C % num_groups == 0
        channels_per_group = C // num_groups
        correlation = (t1 * t2).view([B, num_groups, channels_per_group, D, H, W]).mean(dim=2)
        assert correlation.shape == (B, num_groups, D, H, W)
    elif (len(t1.shape) == 4):
        B, C, H, W = t1.shape
        assert C % num_groups == 0
        channels_per_group = C // num_groups
        correlation = (t1 * t2).view([B, num_groups, channels_per_group, H, W]).mean(dim=2)
        assert correlation.shape == (B, num_groups, H, W)
    else:
        print("Can only compute GWC with 4 or 5 dimension tensors")
        sys.exit()

    return correlation


def non_zero_std(maps: torch.Tensor, device: str, dim: int = 1, keepdim: bool = False) -> torch.Tensor:
    """Computes the standard deviation of all non-zero values in an input Tensor along the given dimension.

    Parameters:
        maps:
        device:
        keepdim:

    Returns:
        The standard deviation of the non-zero elements of the input map.
    """
    batch_size, views, height, width = maps.shape
    valid_map = torch.ne(maps, 0.0).to(torch.float32).to(device)
    valid_count = torch.sum(valid_map, dim=1, keepdim=keepdim)+1e-7
    mean = torch.div(torch.sum(maps,dim=1, keepdim=keepdim), valid_count).reshape(batch_size, 1, height, width).repeat(1,views,1,1)
    mean = torch.mul(valid_map, mean)

    std = torch.sub(maps, mean)
    std = torch.square(std)
    std = torch.sum(std, dim=1, keepdim=keepdim)
    std = torch.div(std, valid_count)
    std = torch.sqrt(std)

    return std


def _normalize_image(image, mean=None, std=None):
    image = image / 255.0

    if mean != None and std != None:
        image = (image - mean) / std

    return image

def normalize(image, mean=None, std=None):
    image = (image-image.min()) / (image.max()-image.min()+1e-10)

    if mean != None and std != None:
        image = (image - mean) / std

    return image

def parameters_count(net, name, do_print=True):
    """

    Parameters:

    Returns:
    """
    model_parameters = filter(lambda p: p.requires_grad, net.parameters())
    params = sum([np.prod(p.size()) for p in model_parameters])
    if do_print:
        print(f"#params {name}: {(params/1e6):0.3f} M")
    return params

def print_gpu_mem() -> None:
    """Prints the current unallocated memory of the GPU.
    """
    t = torch.cuda.get_device_properties(0).total_memory
    r = torch.cuda.memory_reserved(0)
    a = torch.cuda.memory_allocated(0)
    f = t- (a+r)
    print("Free: {:0.4f} GB".format(f/(1024*1024*1024)))

def round_nearest(num: float, decimal: int = 0) -> int:
    """Rounds a floating point number to the nearest decimal place.

    Args:
        num: Float to be rounded.
        decimal: Decimal place to round to.

    Returns:
        The given number rounded to the nearest decimal place.

    Examples:
        >>> round_nearest(11.1)
        11
        >>> round_nearest(15.7)
        16
        >>> round_nearest(2.5)
        2
        >>> round_nearest(3.5)
        3
        >>> round_nearest(14.156, 1)
        14.2
        >>> round_nearest(15.156, 1)
        15.2
        >>> round_nearest(15.156, 2)
        15.16
    """

    return np.round(num+10**(-len(str(num))-1), decimal)

def scale_camera(cam: np.ndarray, scale: float = 1.0) -> np.ndarray:
    """Scales a camera intrinsic parameters.

    Parameters:
        cam: Input camera to be scaled.
        scale: Scale factor.

    Returns:
        The scaled camera.
    """
    new_cam = np.copy(cam)
    new_cam[1][0][0] = cam[1][0][0] * scale
    new_cam[1][1][1] = cam[1][1][1] * scale
    new_cam[1][0][2] = cam[1][0][2] * scale
    new_cam[1][1][2] = cam[1][1][2] * scale
    return new_cam


def scale_image(image: np.ndarray, scale: float = 1.0, interpolation: str = "linear") -> np.ndarray:
    """Scales an input pixel grid.

    Parameters:
        image: Input image to be scaled.
        scale: Scale factor.
        interpolation: Interpolation technique to be used.

    Returns:
        The scaled image.
    """
    if interpolation == 'linear':
        return cv2.resize(image, None, fx=scale, fy=scale, interpolation=cv2.INTER_LINEAR)
    if interpolation == 'nearest':
        return cv2.resize(image, None, fx=scale, fy=scale, interpolation=cv2.INTER_NEAREST)


def scale_mvs_data(depths: np.ndarray, confs: np.ndarray, cams: np.ndarray, scale: float = 1.0, interpolation: str = "linear") -> Tuple[np.ndarray,np.ndarray,np.ndarray]:
    """Scales input depth maps, confidence maps, and cameras.

    Parameters:
        depths: Input depth maps to be scaled.
        confs: Input confidence maps to be scaled.
        cams: Input cameras to be scaled
        scale: Scale factor.
        interpolation: Interpolation technique.

    Returns:
        scaled_depths: The scaled depth maps.
        scaled_confs: The scaled confidence maps.
        cams: The scaled cameras.
    """
    views, height, width = depths.shape

    scaled_depths = []
    scaled_confs = []

    for view in range(views):
        scaled_depths.append(scale_image(depths[view], scale=scale, interpolation=interpolation))
        scaled_confs.append(scale_image(confs[view], scale=scale, interpolation=interpolation))
        cams[view] = scale_camera(cams[view], scale=scale)

    return np.asarray(scaled_depths), np.asarray(scaled_confs), cams


def set_random_seed(seed):
    """

    Parameters:

    Returns:
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)


def to_gpu(data: dict, device: str) -> None:
    """Loads a dictionary of elements onto the GPU device.

    Parameters:
        data: Dictionary to be loaded.
        device: GPU device identifier.
    """

    for key,val in data.items():
        if isinstance(val, torch.Tensor):
            data[key] = val.cuda(device, non_blocking=True)
        if isinstance(val, dict):
            for k1,v1 in val.items():
                if isinstance(v1, torch.Tensor):
                    data[key][k1] = v1.cuda(device, non_blocking=True)
    return


def top_k_hypothesis_selection(grid, k, prev_hypos, prev_hypo_coords, prev_intervals):
    selected_prob, selected_idx = torch.topk(grid,k=k,dim=2)
    selected_hypos = torch.gather(prev_hypos, dim=2, index=selected_idx)
    selected_intervals = torch.gather(prev_intervals, dim=2, index=selected_idx)
    selected_intervals = selected_intervals/2
    selected_coords = torch.gather(prev_hypo_coords, dim=2, index=selected_idx)
    # subdivide hypos
    upper_new_hypos = selected_hypos+selected_intervals/2
    lower_new_hypos = selected_hypos-selected_intervals/2
    new_hypos = torch.cat((upper_new_hypos,lower_new_hypos),dim=2)
    new_hypos = torch.repeat_interleave(new_hypos,2,dim=3)
    new_hypos = torch.repeat_interleave(new_hypos,2,dim=4)

    # subdivide coords
    upper_new_coords = selected_coords*2+1
    lower_new_coords = selected_coords*2
    new_coords = torch.cat((upper_new_coords,lower_new_coords),dim=2)
    new_coords = torch.repeat_interleave(new_coords,2,dim=3)
    new_coords = torch.repeat_interleave(new_coords,2,dim=4)

    # subdivide intervals
    selected_intervals = torch.cat((selected_intervals,selected_intervals),dim=2) # Dx2
    selected_intervals = torch.repeat_interleave(selected_intervals,2,dim=3) # Hx2
    new_intervals = torch.repeat_interleave(selected_intervals,2,dim=4) # Wx2

    return new_hypos, new_coords, new_intervals



# cvt/visualization/video.py
"""Module including routines for creating videos.

This module includes the following functions:

- `video_from_images(image_files, video_file, frame_rate)` - Creates a video from a set of images.
- `video_from_images2(image1_files, image2_files, video_file, frame_rate, orientation)` - Creates a video from two sets of images, stitching the images together each frame.
"""
import cv2
import os
import sys
from typing import List

def video_from_images(image_files: List[str], video_file: str, frame_rate: int = 15) -> None:
    """Creates a video from a set of images.

    Parameters:
        image_files: List of image files to be stitched into a video.
        video_file: Output video file to be created.
        frame_rate: Desired frame rate of the video.
    """
    frame = cv2.imread(image_files[0])
    height, width, layers = frame.shape

    video = cv2.VideoWriter(video_file, 0, fps=frame_rate, frameSize=(width,height))

    for image_file in image_files:
        video.write(cv2.imread(image_file))

    cv2.destroyAllWindows()
    video.release()

    return

def video_from_images2(image1_files: List[str], image2_files: List[str], video_file: str, frame_rate: int = 15, orientation: str = "horizontal") -> None:
    """Creates a video from two sets of images, stitching the images together each frame.

    Parameters:
        image1_files: List of first image files to be stitched.
        image2_files: List of second image files to be stitched.
        video_file: Output video file to be created.
        frame_rate: Desired frame rate of the video.
        orientation: Desired stitch orientation (horizontal or verticle).
    """
    frame1 = cv2.imread(image1_files[0])
    height1, width1, layers1 = frame1.shape

    frame2 = cv2.imread(image2_files[0])
    height2, width2, layers2 = frame2.shape

    assert(len(image1_files) == len(image2_files))
    assert(height1==height2)
    assert(width1==width2)
    assert(layers1==layers2)

    if (orientation=="horizontal"):
        video = cv2.VideoWriter(video_file, 0, fps=frame_rate, frameSize=((width*2,height)))
        for (img1_file, img2_file) in zip(image1_files, image2_files):
            img1 = cv2.imread(img1_file)
            img2 = cv2.imread(img2_file)

            frame = np.concatenate((img1,img2), axis=1)
            video.write(frame)
        cv2.destroyAllWindows()
        video.release()

    else:
        print(f"Orientation '{orientation}' is not yet supported.")

    return




# cvt/geometry.py

"""Module including geometric routines.

This module contains the following functions:

- `downsample_cloud(cloud, min_point_dist)` - Downsamples a point cloud enforcing a minumum point spacing.
- `essential_from_features(src_image_file, tgt_image_file, K)` - Computes the essential matrix between two images using image features.
- `fundamental_from_KP(K, P_src, P_tgt)` - Computes the fundamental matrix between two images using camera parameters.
- `fundamental_from_features(src_image_file, tgt_image_file)` - Computes the fundamental matrix between two images using image features.
- `geometric_consistency_error(src_depth, src_cam, tgt_depth, tgt_cam)` - .
- `geometric_consistency_mask(src_depth, src_cam, tgt_depth, tgt_cam, pixel_th)` - Computes the geometric consistency mask between a source and target depth map.
- `homography(src_image_file, tgt_image_file)` - Computes a homography transformation between two images using image features.
- `homography_warp(cfg,features, level, ref_in, src_in, ref_ex, src_ex, depth_hypos, group_channels, vwa_net, vis_weights, aggregation)` - Performs homography warping to create a Plane Sweeping Volume (PSV).
- `match_features(src_image, tgt_image, max_features)` - Computer matching ORB features between a pair of images.
- `plane_coords(K, P, depth_hypos, H, W)` - .
- `points_from_depth(depth, cam)` - Creates a point array from a single depth map.
- `project_depth_map(depth, cam, mask)` - Projects a depth map into a list of 3D points.
- `project_renderer(renderer, K, P, width, height)` - Projects the scene in an Open3D Offscreen Renderer to the 2D image plane.
- `render_custom_values(points, values, image_shape, cam)` - Renders a point cloud into a 2D camera plane using custom values for each pixel.
- `render_point_cloud(cloud, cam, width, height)` - Renders a point cloud into a 2D image plane.
- `reproject(src_depth, src_cam, tgt_depth, tgt_cam)` - Computes the re-projection depth values and pixel indices between two depth maps.
- `sample_volume(volume, z_vals, coords, H, W, near_depth, far_depth, inv_depth)` - .
- `soft_hypothesis(data, target_hypo, focal_length, min_hypo, max_hypo, M, delta_in)` - .
- `visibility(depths, K, Ps, vis_th, levels)` - .
- `visibility_mask(src_depth, src_cam, depth_files, cam_files, src_ind, pixel_th)` - Computes a visibility mask between a provided source depth map and list of target depth maps.
- `uniform_hypothesis(cfg, ref_in, src_in, ref_ex, src_ex, depth_min, depth_max, img_height, img_width, nhypothesis_init, inv_depth)` - .
"""

import numpy as np
import sys
import cv2
import open3d as o3d
from typing import Tuple, List, Optional
import torch
import torch.nn.functional as F
import torchvision.transforms.functional as tvf
import torchvision.transforms as tvt
from torch.cuda.amp import autocast

import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle

from camera import intrinsic_pyramid, Z_from_disp
from common import groupwise_correlation
from io import *

def downsample_cloud(cloud, min_point_dist):
    """Downsamples a point cloud enforcing a minumum point spacing.
    Parameters:
        cloud: Point cloud to be decimated.
        min_point_dist: minimum point spacing to enforce.

    Returns:
        The downsampled point cloud.
    """
    return cloud.voxel_down_sample(voxel_size=min_point_dist)

def edge_mask(depth, near_depth, far_depth):
    down_gt = F.interpolate(depth,scale_factor=0.5,mode='bilinear',align_corners=False,recompute_scale_factor=False)
    down_up_gt = F.interpolate(down_gt,scale_factor=2,mode='bilinear',align_corners=False,recompute_scale_factor=False)
    res = torch.abs(depth-down_up_gt)
    high_frequency_mask = res>(0.001*(far_depth-near_depth)[:,None,None,None])
    valid_gt_mask = (-F.max_pool2d(-depth,kernel_size=5,stride=1,padding=2))>near_depth[:,None,None,None]
    high_frequency_mask = high_frequency_mask * valid_gt_mask
    high_frequency_mask = ((1-high_frequency_mask.to(torch.int32)) * valid_gt_mask).to(torch.int32)
    return high_frequency_mask

def get_epipolar_inds(x0, y0, x1, y1, x_lim, y_lim, max_patches):
    dx = x1-x0
    dy = y1-y0
    yi = torch.ones_like(y0)

    negative_mask = torch.where(dy < 0, -1, 1)
    yi *= negative_mask
    dy *= negative_mask

    D = (2*dy) - dx
    y = y0
    x = x0

    batch,h,w = x.shape

    epipolar_grid = torch.zeros((batch, max_patches, h, w, 2)).to(x0)
    for i in range(max_patches):
        # build valid indices mask
        valid_mask = torch.where((x < x_lim).to(torch.bool) & (x >= 0).to(torch.bool), 1, 0)
        valid_mask *= torch.where((y < y_lim).to(torch.bool) & (y >= 0).to(torch.bool), 1, 0)
        valid_mask = valid_mask.unsqueeze(-1).repeat(1,1,1,2)
        
        # stack xy and apply valid indices mask
        xy = torch.stack([x,y],dim=-1)
        epipolar_grid[:,i,:,:,:] = (xy*valid_mask) - (1-valid_mask)

        mask = torch.where(D > 0, 1, 0)
        y = (y+yi)*mask + y*(1-mask)
        D = ((D + (2*(dy-dx)))*mask) + ((D + (2*dy))*(1-mask))
        x += 1

    return epipolar_grid[:,:,:,:,0], epipolar_grid[:,:,:,:,1]

def epipolar_patch_retrieval(imgs, intrinsics, extrinsics, patch_size):
    batch_size, views, c, height, width = imgs.shape
    half_patch_size = patch_size//2

    x_flat = torch.arange((half_patch_size),width+1)[::patch_size].to(imgs)
    y_flat = torch.arange((half_patch_size),height+1)[::patch_size].to(imgs)

    xgrid, ygrid = torch.meshgrid([x_flat,y_flat], indexing="xy")
    xy = torch.stack([xgrid, ygrid, torch.ones_like(xgrid)], dim=-1) # [patched_height, patch_width, 3]
    patched_height, patched_width, _ = xy.shape
    xy = xy.unsqueeze(0).repeat(batch_size, 1, 1, 1).unsqueeze(-1) # [batch_size, patched_height, patch_width, 3, 1]

    max_patches = patched_height+patched_width-1

    patch_volumes = []
    for j in range(imgs.shape[1]):
        #view_patches = [imgs[:,j].unsqueeze(2)]
        view_patches = []

        K = intrinsics[:,j]
        P_src = extrinsics[:,j]

        for i in range(imgs.shape[1]):
            if i==j:
                continue

            x_lim = (torch.ones((batch_size, patched_height, patched_width)) * patched_width).to(imgs)
            y_lim = (torch.ones((batch_size, patched_height, patched_width)) * patched_height).to(imgs)
        
            P_tgt = extrinsics[:,i]
            Fm = fundamental_from_KP(K, P_src, P_tgt)
            Fm = Fm.reshape(batch_size, 1, 1, 3, 3).repeat(1, xy.shape[1], xy.shape[2], 1, 1) # [batch_size, patched_height, patch_width, 3, 1]
            line = torch.matmul(Fm,xy).squeeze(-1) # [batch_size, patched_height, patch_width, 3]
            
            ## Start Point ##
            # initial x coordinate, comput y corrdinate
            x0 = x_flat[0].reshape(1,1,1).repeat(batch_size, patched_height, patched_width)
            y0 = (-(line[:,:,:,0]/line[:,:,:,1])*x0) - (line[:,:,:,2]/line[:,:,:,1])
            # check for invalid y coordinates
            y_mask_lt = torch.where(y0 < 0, 1, 0)
            y_mask_gt = torch.where(y0 >= height, 1, 0)
            y_mask_out = y_mask_lt+y_mask_gt
            # adjust for invalid y coordinates
            y0 = y0*(1-y_mask_lt) + y_flat[0]*y_mask_lt
            y0 = y0*(1-y_mask_gt) + y_flat[-1]*y_mask_gt
            x0 = (x0 * (1-y_mask_out)) + (((-(line[:,:,:,1]/line[:,:,:,0])*y0) - (line[:,:,:,2]/line[:,:,:,0])) * (y_mask_out))
            # if x coordinate is invalid
            valid_mask = torch.where((x0 >= 0).to(torch.bool) & (x0 < width).to(torch.bool), 1, 0)
            x0 = (x0*valid_mask) - (1-valid_mask)
            y0 = (y0*valid_mask) - (1-valid_mask)

            ## End Point ##
            # initial x coordinate, comput y corrdinate
            x1 = x_flat[-1].reshape(1,1,1).repeat(batch_size, patched_height, patched_width)
            y1 = (-(line[:,:,:,0]/line[:,:,:,1])*x1) - (line[:,:,:,2]/line[:,:,:,1])
            # check for invalid y coordinates
            y_mask_lt = torch.where(y1 < 0, 1, 0)
            y_mask_gt = torch.where(y1 >= height, 1, 0)
            y_mask_out = y_mask_lt+y_mask_gt
            # adjust for invalid y coordinates
            y1 = y1*(1-y_mask_lt) + y_flat[0]*y_mask_lt
            y1 = y1*(1-y_mask_gt) + y_flat[-1]*y_mask_gt
            x1 = (x1 * (1-y_mask_out)) + (((-(line[:,:,:,1]/line[:,:,:,0])*y1) - (line[:,:,:,2]/line[:,:,:,0])) * (y_mask_out))
            # if x coordinate is invalid
            valid_mask = torch.where((x1 >= 0).to(torch.bool) & (x1 < width).to(torch.bool), 1, 0)
            x1 = (x1*valid_mask) - (1-valid_mask)
            y1 = (y1*valid_mask) - (1-valid_mask)

            # convert image indices into patch indices
            x0 = torch.round((x0-(half_patch_size))/patch_size)
            x1 = torch.round((x1-(half_patch_size))/patch_size)
            y0 = torch.round((y0-(half_patch_size))/patch_size)
            y1 = torch.round((y1-(half_patch_size))/patch_size)

            # compute x and y slopes
            slope_x = torch.abs(x1-x0)
            slope_y = torch.abs(y1-y0)

            # flip x's and y's depending on slope
            small_slope_mask = torch.where(slope_y < slope_x, 1, 0)
            x0_temp = x0*small_slope_mask + y0*(1 - small_slope_mask)
            y0 = y0*small_slope_mask + x0*(1 - small_slope_mask)
            x0 = x0_temp
            x1_temp = x1*small_slope_mask + y1*(1 - small_slope_mask)
            y1 = y1*small_slope_mask + x1*(1 - small_slope_mask)
            x1 = x1_temp
            x_lim_temp = x_lim*small_slope_mask + y_lim*(1 - small_slope_mask)
            y_lim = y_lim*small_slope_mask + x_lim*(1 - small_slope_mask)
            x_lim = x_lim_temp
            
            # flip start and end points so start is smaller
            small_end_mask = torch.where(x1 < x0, 1, 0)
            x0_temp = x0*(1-small_end_mask) + x1*small_end_mask
            x1 = x1*(1-small_end_mask) + x0*small_end_mask
            x0 = x0_temp
            y0_temp = y0*(1-small_end_mask) + y1*small_end_mask
            y1 = y1*(1-small_end_mask) + y0*small_end_mask
            y0 = y0_temp

            # grab nearest patch indices
            x_grid, y_grid = get_epipolar_inds(x0, y0, x1, y1, x_lim, y_lim, max_patches)

            # flip x and y indices back where necessary (using small_slope_mask)
            small_slope_mask = small_slope_mask.reshape(batch_size,1,patched_height,patched_width).repeat(1,max_patches,1,1)
            x_grid_temp = x_grid*small_slope_mask + y_grid*(1 - small_slope_mask)
            y_grid = y_grid*small_slope_mask + x_grid*(1 - small_slope_mask)
            x_grid = x_grid_temp
            epipolar_grid = torch.stack([x_grid,y_grid], dim=-1)

            # convert patch indices into image indices
            epipolar_grid = epipolar_grid*patch_size + (half_patch_size)
            epipolar_grid = torch.where(epipolar_grid < 0, -1, epipolar_grid)
            patch_grid = epipolar_grid.cpu().numpy()

            # duplicate patch center indices over entire patch
            epipolar_grid = torch.repeat_interleave(epipolar_grid, patch_size, dim=2)
            epipolar_grid = torch.repeat_interleave(epipolar_grid, patch_size, dim=3)

            # apply center offset matrix
            valid_mask = torch.where(epipolar_grid >= 0, 1, 0)
            patch_offset = torch.arange(-half_patch_size,half_patch_size).to(imgs)
            x_offset, y_offset = torch.meshgrid([patch_offset,patch_offset], indexing="xy")
            patch_offset = torch.stack([x_offset,y_offset],dim=-1)
            patch_offset = torch.tile(patch_offset, (patched_height, patched_width,1))
            patch_offset = patch_offset.reshape(1,1,height,width,2).repeat(batch_size,max_patches,1,1,1)
            epipolar_grid += patch_offset
            epipolar_grid = (epipolar_grid * valid_mask) - (1-valid_mask)

            # normalize coordinate grid
            min_coord = torch.tensor([0,0]).to(imgs)
            min_coord = min_coord.reshape(1,1,1,1,2).repeat(batch_size, max_patches, height, width, 1)
            max_coord = torch.tensor([width-1,height-1]).to(imgs)
            max_coord = max_coord.reshape(1,1,1,1,2).repeat(batch_size, max_patches, height, width, 1)
            norm_grid = (epipolar_grid - min_coord) / (max_coord - min_coord)
            norm_grid = (norm_grid * 2) - 1

            # aggregate image patches
            img_patches = F.grid_sample(imgs[:,i],
                                        norm_grid.reshape(batch_size, max_patches, height*width, 2),
                                        mode="nearest",
                                        padding_mode="zeros")
            img_patches = img_patches.reshape(batch_size, 3, max_patches, height, width)
            view_patches.append(img_patches)

            #   #### visual
            #   r,c = patched_height//2, patched_width//2
            #   #for k in range(max_patches):
            #   #    x_k,y_k,_ = xy[0,r,c,:,0].cpu().numpy()
            #   #    fig = plt.figure()
            #   #    ax = fig.add_subplot(111)
            #   #    ax.imshow(torch.movedim(img_patches[0,:,k], (0,1,2), (2,0,1)).cpu().numpy())
            #   #    rect_k = Rectangle((x_k-(half_patch_size),y_k-(half_patch_size)), patch_size, patch_size, color='red', fc = 'none', lw = 0.5)
            #   #    ax.add_patch(rect_k)
            #   #    plt.savefig(f"patches/ref{j:02d}_src{i:02d}_patch{k:04d}.png")
            #   #    plt.close()

            #   # plot src patches
            #   ep = patch_grid[0,:,r,c,:]
            #   fig = plt.figure()
            #   ax = fig.add_subplot(111)
            #   ax.imshow(torch.movedim(imgs[0,i],(0,1,2),(2,0,1)).cpu().numpy())
            #   for x_i,y_i in ep:
            #       if x_i>=0 and y_i>=0:
            #           rect_i = Rectangle((x_i-(half_patch_size),y_i-(half_patch_size)), patch_size, patch_size, color='red', fc = 'none', lw = 0.5)
            #           ax.add_patch(rect_i)
            #   plt.axis('off')
            #   plt.savefig(f"patches/ref{j:02d}_src{i:02d}.png", bbox_inches='tight', dpi=300)
            #   plt.close()
            #   #### visual
        #   #### visual
        #   # plot ref point
        #   ref_pix = xy[0,r,c,:,0].cpu().numpy()
        #   plt.imshow(torch.movedim(imgs[0,j],(0,1,2),(2,0,1)).cpu().numpy())
        #   plt.plot(ref_pix[0].item(),ref_pix[1].item(),'ro')
        #   plt.axis('off')
        #   plt.savefig(f"patches/ref{j:02d}.png", bbox_inches='tight', dpi=300)
        #   plt.close()
        #   sys.exit()
        #   #### visual

        patch_volumes.append(torch.cat(view_patches, dim=2))
    return torch.stack(patch_volumes, dim=1)

def essential_from_features(src_image_file: str, tgt_image_file: str, K: np.ndarray) -> np.ndarray:
    """Computes the essential matrix between two images using image features.

    Parameters:
        src_image_file: Input file for the source image.
        tgt_image_file: Input file for the target image.
        K: Intrinsics matrix of the two cameras (assumed to be constant between views).

    Returns:
        The essential matrix betweent the two image views.
    """
    src_image = cv2.imread(src_image_file)
    tgt_image = cv2.imread(tgt_image_file)

    # compute matching features
    (src_points, tgt_points) = match_features(src_image, tgt_image)

    # Compute fundamental matrix
    E, mask = cv2.findEssentialMat(src_points, tgt_points, K, method=cv2.RANSAC)

    return E

def fundamental_from_KP(K: np.ndarray, P_src: np.ndarray, P_tgt: np.ndarray) -> np.ndarray:
    """Computes the fundamental matrix between two images using camera parameters.

    Parameters:
        K: Intrinsics matrix of the two cameras (assumed to be constant between views).
        P_src: Extrinsics matrix for the source view.
        P_tgt: Extrinsics matrix for the target view.

    Returns:
        The fundamental matrix betweent the two cameras.
    """
    F_mats = []
    for i in range(K.shape[0]):
        R1 = P_src[i,0:3,0:3]
        t1 = P_src[i,0:3,3]
        R2 = P_tgt[i,0:3,0:3]
        t2 = P_tgt[i,0:3,3]

        t1aug = torch.tensor([t1[0], t1[1], t1[2], 1]).to(K)
        epi2 = torch.matmul(P_tgt[i],t1aug)
        epi2 = torch.matmul(K[i],epi2[0:3])

        R = torch.matmul(R2,torch.t(R1))
        t = t2 - torch.matmul(R,t1)
        K1inv = torch.linalg.inv(K[i])
        K2invT = torch.t(K1inv)
        tx = torch.tensor([[0, -t[2], t[1]], [t[2], 0, -t[0]], [-t[1], t[0], 0]]).to(K)
        F = torch.matmul(K2invT,torch.matmul(tx,torch.matmul(R,K1inv)))
        F = F/(torch.max(F)+1e-10)
        F_mats.append(F)

    return torch.stack(F_mats, dim=0)

def _fundamental_from_KP(K: np.ndarray, P_src: np.ndarray, P_tgt: np.ndarray) -> np.ndarray:
    """Computes the fundamental matrix between two images using camera parameters.

    Parameters:
        K: Intrinsics matrix of the two cameras (assumed to be constant between views).
        P_src: Extrinsics matrix for the source view.
        P_tgt: Extrinsics matrix for the target view.

    Returns:
        The fundamental matrix betweent the two cameras.
    """
    R1 = P_src[0:3,0:3]
    t1 = P_src[0:3,3]
    R2 = P_tgt[0:3,0:3]
    t2 = P_tgt[0:3,3]

    t1aug = np.array([t1[0], t1[1], t1[2], 1])
    epi2 = np.matmul(P_tgt,t1aug)
    epi2 = np.matmul(K,epi2[0:3])

    R = np.matmul(R2,np.transpose(R1))
    t= t2- np.matmul(R,t1)
    K1inv = np.linalg.inv(K)
    K2invT = np.transpose(K1inv)
    tx = np.array([[0, -t[2], t[1]], [t[2], 0, -t[0]], [-t[1], t[0], 0]])
    F = np.matmul(K2invT,np.matmul(tx,np.matmul(R,K1inv)))
    F = F/np.amax(F)

    return F

def fundamental_from_features(src_image_file: str, tgt_image_file: str) -> np.ndarray:
    """Computes the fundamental matrix between two images using image features.

    Parameters:
        src_image_file: Input file for the source image.
        tgt_image_file: Input file for the target image.

    Returns:
        The fundamental matrix betweent the two image views.
    """
    src_image = cv2.imread(src_image_file)
    tgt_image = cv2.imread(tgt_image_file)

    # compute matching features
    (src_points, tgt_points) = match_features(src_image, tgt_image)
    
    # Compute fundamental matrix
    F, mask = cv2.findFundamentalMat(src_points,tgt_points,cv2.FM_8POINT)

    return F

def geometric_consistency_error(src_depth: np.ndarray, src_cam: np.ndarray, tgt_depth: np.ndarray, tgt_cam: np.ndarray) -> np.ndarray:
    """Computes the geometric consistency error between a source and target depth map.

    Parameters:
        src_depth: Depth map for the source view.
        src_cam: Camera parameters for the source depth map viewpoint.
        tgt_depth: Depth map for the target view.
        tgt_cam: Camera parameters for the target depth map viewpoint.

    Returns:
        The binary consistency mask encoding depth consensus between source and target depth maps.
    """
    height, width = src_depth.shape
    x_src, y_src = np.meshgrid(np.arange(0, width), np.arange(0, height))

    depth_reprojected, coords_reprojected, coords_tgt, projection_map = reproject(src_depth, src_cam, tgt_depth, tgt_cam)

    dist = np.sqrt((coords_reprojected[:,:,0] - x_src) ** 2 + (coords_reprojected[:,:,1] - y_src) ** 2)

    return dist, projection_map

def geometric_consistency_mask(src_depth, src_K, src_P, tgt_depth, tgt_K, tgt_P, pixel_th):
    """Computes the geometric consistency mask between a source and target depth map.

    Parameters:
        src_depth: Depth map for the source view.
        src_K: Intrinsic camera parameters for the source depth map viewpoint.
        src_P: Extrinsic camera parameters for the source depth map viewpoint.
        tgt_depth: Target depth map used for re-projection.
        tgt_K: Intrinsic camera parameters for the target depth map viewpoint.
        tgt_P: Extrinsic camera parameters for the target depth map viewpoint.
        pixel_th: Pixel re-projection threshold to determine matching depth estimates.

    Returns:
        The binary consistency mask encoding depth consensus between source and target depth maps.
    """
    batch_size, c, height, width = src_depth.shape
    depth_reprojected, coords_reprojected, coords_tgt = reproject(src_depth, src_K, src_P, tgt_depth, tgt_K, tgt_P)

    x_src, y_src = torch.meshgrid(torch.arange(0, width), torch.arange(0, height), indexing="xy")
    x_src = x_src.unsqueeze(0).repeat(batch_size, 1, 1).to(src_depth)
    y_src = y_src.unsqueeze(0).repeat(batch_size, 1, 1).to(src_depth)
    dist = torch.sqrt((coords_reprojected[:,:,:,0] - x_src) ** 2 + (coords_reprojected[:,:,:,1] - y_src) ** 2)

    mask = torch.where(dist < pixel_th, 1, 0)
    return mask

def _geometric_consistency_mask(src_depth, src_K, src_P, tgt_depth, tgt_K, tgt_P, pixel_th):
    """Computes the geometric consistency mask between a source and target depth map.

    Parameters:
        src_depth: Depth map for the source view.
        src_K: Intrinsic camera parameters for the source depth map viewpoint.
        src_P: Extrinsic camera parameters for the source depth map viewpoint.
        tgt_depth: Target depth map used for re-projection.
        tgt_K: Intrinsic camera parameters for the target depth map viewpoint.
        tgt_P: Extrinsic camera parameters for the target depth map viewpoint.
        pixel_th: Pixel re-projection threshold to determine matching depth estimates.

    Returns:
        The binary consistency mask encoding depth consensus between source and target depth maps.
    """
    height, width = src_depth.shape
    x_src, y_src = np.meshgrid(np.arange(0, width), np.arange(0, height))

    depth_reprojected, coords_reprojected, coords_tgt = _reproject(src_depth, src_K, src_P, tgt_depth, tgt_K, tgt_P)

    dist = np.sqrt((coords_reprojected[:,:,0] - x_src) ** 2 + (coords_reprojected[:,:,1] - y_src) ** 2)
    mask = np.where(dist < pixel_th, 1, 0)

    return mask

def get_uncovered_mask(data, output):
    hypos = output["hypos"]
    intervals = output["intervals"]
    _,H,W = data["target_depth"].shape
    levels = len(hypos)

    uncovered_masks = torch.zeros(levels,H,W)
    for l in range(levels):
        hypo = hypos[l][0].squeeze(0)
        planes, h, w = hypo.shape
        interval = intervals[l][0].squeeze(0)
        target_depth = tvf.resize(data["target_depth"][0:1].unsqueeze(0), [h,w]).reshape(h,w)

        ### compute coverage
        diff = torch.abs(hypo - target_depth)
        min_interval = interval[:,0:1] * 0.5 # intervals are bin widths, divide by 2 for radius
        coverage = torch.clip(torch.where(diff <= min_interval, 1, 0).sum(dim=0, keepdim=True), 0, 1)
        uncovered = torch.clip(torch.where(coverage <= 0, 1, 0).sum(dim=0, keepdim=True), 0, 1)
        valid_targets = torch.where(target_depth > 0, 1, 0)
        uncovered *= valid_targets
        uncovered_masks[l] = (tvf.resize(uncovered.reshape(1,1,h,w), [H,W], interpolation=tvt.InterpolationMode.NEAREST)).reshape(H,W)

    return uncovered_masks

def homography(src_image_file: str, tgt_image_file: str) -> np.ndarray:
    """Computes a homography transformation between two images using image features.

    Parameters:
        src_image_file: Input file for the source image.
        tgt_image_file: Input file for the target image.

    Returns:
        The homography matrix to warp the target image to the source image.
    """
    src_image = cv2.imread(src_image_file)
    tgt_image = cv2.imread(tgt_image_file)

    (height, width, _) = src_image.shape

    (src_points, tgt_points) = match_features(src_image, tgt_image)

    # Compute fundamental matrix
    H, mask = cv2.findHomography(tgt_points, src_points, method=cv2.RANSAC)

    return H

def psv(cfg, images, intrinsics, extrinsics, depth_hypos):
    """Performs homography warping to create a Plane Sweeping Volume (PSV).
    Parameters:
        cfg: Configuration dictionary containing configuration parameters.
        images: image maps to be warped into a PSV.
        intrinsics: intrinsics matrices.
        extrinsics: extrinsics matrices.
        depth_hypos: Depth hypotheses to use for homography warping.

    Returns:
        The Plane Sweeping Volume computed via feature matching cost.
    """
    depth_hypos = depth_hypos.squeeze(1)
    _,planes,_,_ = depth_hypos.shape
    B,views,C,H,W = images.shape

    pairwise_psv = []
    for v in range(1,views):
        with torch.no_grad():
            src_proj = torch.matmul(intrinsics[:,v], extrinsics[:,v,0:3])
            ref_proj = torch.matmul(intrinsics[:,0], extrinsics[:,0,0:3])
            last = torch.tensor([[[0,0,0,1.0]]]).repeat(B,1,1).to(images.device)
            src_proj = torch.cat((src_proj,last),1)
            ref_proj = torch.cat((ref_proj,last),1)

            proj = torch.matmul(src_proj, torch.inverse(ref_proj))
            rot = proj[:, :3, :3]  # [B,3,3]
            trans = proj[:, :3, 3:4]  # [B,3,1]

            y, x = torch.meshgrid([torch.arange(0, H, dtype=torch.float32, device=images.device),
                                torch.arange(0, W, dtype=torch.float32, device=images.device)],
                                indexing='ij')
            y, x = y.contiguous(), x.contiguous()
            y, x = y.view(H * W), x.view(H * W)
            xyz = torch.stack((x, y, torch.ones_like(x)))  # [3, H*W]
            xyz = torch.unsqueeze(xyz, 0).repeat(B, 1, 1)  # [B, 3, H*W]
            rot_xyz = torch.matmul(rot, xyz)  # [B, 3, H*W]

            rot_depth_xyz = rot_xyz.unsqueeze(2).repeat(1, 1, planes, 1) * depth_hypos.view(B, 1, planes, H*W)  # [B, 3, Ndepth, H*W]
            proj_xyz = rot_depth_xyz + trans.view(B, 3, 1, 1)  # [B, 3, Ndepth, H*W]
            proj_xy = proj_xyz[:, :2, :, :] / proj_xyz[:, 2:3, :, :]  # [B, 2, Ndepth, H*W]
            proj_x_normalized = proj_xy[:, 0, :, :] / ((W - 1) / 2) - 1
            proj_y_normalized = proj_xy[:, 1, :, :] / ((H - 1) / 2) - 1
            proj_xy = torch.stack((proj_x_normalized, proj_y_normalized), dim=3)  # [B, Ndepth, H*W, 2]
            grid = proj_xy
        grid = grid.type(images.dtype)

        warped_src_image = F.grid_sample( images[:,v],
                                        grid.view(B, planes * H, W, 2), 
                                        mode='bilinear',
                                        padding_mode='zeros',
                                        align_corners=False)
        pairwise_psv.append(warped_src_image.view(B, C, planes, H, W))

    return pairwise_psv

def homography_warp_var(cfg, features, ref_in, src_in, ref_ex, src_ex, depth_hypos):
    """Performs homography warping to create a Plane Sweeping Volume (PSV).
    Parameters:
        cfg: Configuration dictionary containing configuration parameters.
        features: Feature maps to be warped into a PSV.
        level: Current feature resolution level.
        ref_in: Reference view intrinsics matrix.
        src_in: Source view intrinsics matrices.
        ref_ex: Reference view extrinsics matrix.
        src_ex: Source view extrinsics matrices.
        depth_hypos: Depth hypotheses to use for homography warping.

    Returns:
        The Plane Sweeping Volume computed via feature matching cost.
    """
    depth_hypos = depth_hypos.squeeze(1)
    _,planes,_,_ = depth_hypos.shape

    B,fCH,H,W = features[0].shape
    num_depth = depth_hypos.shape[1]
    nSrc = len(features)-1

    vis_weight_list = []
    ref_volume = features[0].unsqueeze(2).repeat(1,1,num_depth,1,1)

    cost_volume = torch.zeros((nSrc+1,B,fCH,planes,H,W)).to(features[0])
    cost_volume[0] = ref_volume
    reweight_sum = None
    for src in range(nSrc):
        with torch.no_grad():
            with autocast(enabled=False):
                src_proj = torch.matmul(src_in[:,src,:,:],src_ex[:,src,0:3,:])
                ref_proj = torch.matmul(ref_in,ref_ex[:,0:3,:])
                last = torch.tensor([[[0,0,0,1.0]]]).repeat(len(src_in),1,1).cuda()
                src_proj = torch.cat((src_proj,last),1)
                ref_proj = torch.cat((ref_proj,last),1)

                proj = torch.matmul(src_proj, torch.inverse(ref_proj))
                rot = proj[:, :3, :3]  # [B,3,3]
                trans = proj[:, :3, 3:4]  # [B,3,1]

                y, x = torch.meshgrid([torch.arange(0, H, dtype=torch.float32, device=ref_volume.device),
                                    torch.arange(0, W, dtype=torch.float32, device=ref_volume.device)],
                                    indexing='ij')
                y, x = y.contiguous(), x.contiguous()
                y, x = y.view(H * W), x.view(H * W)
                xyz = torch.stack((x, y, torch.ones_like(x)))  # [3, H*W]
                xyz = torch.unsqueeze(xyz, 0).repeat(B, 1, 1)  # [B, 3, H*W]
                rot_xyz = torch.matmul(rot, xyz)  # [B, 3, H*W]

                rot_depth_xyz = rot_xyz.unsqueeze(2).repeat(1, 1, num_depth, 1) * depth_hypos.view(B, 1, num_depth,H*W)  # [B, 3, Ndepth, H*W]
                proj_xyz = rot_depth_xyz + trans.view(B, 3, 1, 1)  # [B, 3, Ndepth, H*W]
                proj_xy = proj_xyz[:, :2, :, :] / proj_xyz[:, 2:3, :, :]  # [B, 2, Ndepth, H*W]
                proj_x_normalized = proj_xy[:, 0, :, :] / ((W - 1) / 2) - 1
                proj_y_normalized = proj_xy[:, 1, :, :] / ((H - 1) / 2) - 1
                proj_xy = torch.stack((proj_x_normalized, proj_y_normalized), dim=3)  # [B, Ndepth, H*W, 2]
                grid = proj_xy

        grid = grid.type(ref_volume.dtype)
        warped_src_fea = F.grid_sample( features[src+1],
                                        grid.view(B, num_depth * H, W, 2), 
                                        mode='bilinear',
                                        padding_mode='zeros',
                                        align_corners=False)
        cost_volume[src+1] = warped_src_fea.view(B, fCH, num_depth, H, W)

        torch.cuda.empty_cache()

    cost_volume = torch.var(cost_volume, dim=0)
    B,C,D,H,W = cost_volume.shape

    return cost_volume


def homography_warp(cfg, features, intrinsics, extrinsics, hypotheses, group_channels, vwa_net=None, vis_weights=None, virtual=False):
    """Performs homography warping to create a Plane Sweeping Volume (PSV).
    Parameters:
        cfg: Configuration dictionary containing configuration parameters.
        features: Feature maps to be warped into a PSV.
        intrinsics: Intrinsics matrices for all views.
        extrinsics: Extrinsics matrices for all views.
        hypotheses: Depth hypotheses to use for homography warping.
        group_channels: Feature channel sizes used in group-wise correlation (GWC).
        vwa_net: Network used for visibility weighting.
        vis_weights: Pre-computed visibility weights.

    Returns:
        The Plane Sweeping Volume computed via feature matching cost.
    """
    hypotheses = hypotheses.squeeze(1)
    _,planes,_,_ = hypotheses.shape
    batch_size, C, height, width = features[0].shape
    num_views = len(features)
    device = features[0].device

    if not virtual:
        ref_volume = features[0].unsqueeze(2).repeat(1,1,planes,1,1)

    vis_weight_list = []
    cost_volume = torch.zeros((batch_size,group_channels,planes,height,width), dtype=torch.float32, device=device)
    reweight_sum = torch.zeros((batch_size,1,planes,height,width), dtype=torch.float32, device=device)

    # build reference projection matrix
    ref_proj = torch.matmul(intrinsics[:,0], extrinsics[:,0,0:3])
    last = torch.tensor([[[0,0,0,1.0]]]).repeat(batch_size, 1, 1).cuda()
    ref_proj = torch.cat((ref_proj, last), 1)
    
    # build coordinates grid
    y, x = torch.meshgrid([torch.arange(0, height, dtype=torch.float32, device=device),
                        torch.arange(0, width, dtype=torch.float32, device=device)],
                        indexing='ij')
    y, x = y.contiguous(), x.contiguous()
    y, x = y.view(height * width), x.view(height * width)
    xyz = torch.stack((x, y, torch.ones_like(x)))
    xyz = torch.unsqueeze(xyz, 0).repeat(batch_size, 1, 1)

    for v in range(1,num_views):
        with torch.no_grad():
            # build source projection matrix
            src_proj = torch.matmul(intrinsics[:,v], extrinsics[:,v,0:3])
            src_proj = torch.cat((src_proj, last), 1)

            # compute full projection matrix between views
            proj = torch.matmul(src_proj, torch.inverse(ref_proj))
            rot = proj[:, :3, :3]
            trans = proj[:, :3, 3:4]

            # Build plane-sweeping coordinates grid between views
            rot_xyz = torch.matmul(rot, xyz)
            rot_depth_xyz = rot_xyz.unsqueeze(2).repeat(1, 1, planes, 1) * hypotheses.view(batch_size, 1, planes, height*width)
            proj_xyz = rot_depth_xyz + trans.view(batch_size, 3, 1, 1)
            proj_xy = proj_xyz[:, :2] / proj_xyz[:, 2:3]
            proj_x_normalized = proj_xy[:, 0] / ((width - 1) / 2) - 1
            proj_y_normalized = proj_xy[:, 1] / ((height - 1) / 2) - 1
            proj_xy = torch.stack((proj_x_normalized, proj_y_normalized), dim=3)
            grid = proj_xy
            grid = grid.type(torch.float32)

        src_feature = features[v]
        warped_features = F.grid_sample( src_feature,
                                        grid.view(batch_size, planes * height, width, 2), 
                                        mode='bilinear',
                                        padding_mode='zeros',
                                        align_corners=False)
        warped_features = warped_features.view(batch_size, C, planes, height, width)

        if virtual:
            if v==1:
                # We continue to the next source view to compute inner product
                # using the first source view is used as reference
                ref_volume = warped_features
                continue

        # compute Pairwise Plane-Sweeping Volume using GWC
        ppsv = groupwise_correlation(warped_features, ref_volume, group_channels)
        if vwa_net is not None:
            reweight = vwa_net(ppsv)
            vis_weight_list.append(reweight)
            reweight = reweight.unsqueeze(1)
            ppsv = reweight*ppsv
        elif vis_weights is not None:
            reweight = vis_weights[v-1]
            if reweight.shape[2] < ppsv.shape[3]:
                reweight = F.interpolate(reweight,scale_factor=2,mode='bilinear',align_corners=False)
            vis_weight_list.append(reweight)
            reweight = reweight.unsqueeze(2)
            ppsv = reweight*ppsv

        cost_volume = cost_volume + ppsv
        reweight_sum = reweight_sum + reweight

        if cfg["mode"]=="inference":
            del src_feature
            del ppsv
            del warped_features
            del reweight
            torch.cuda.empty_cache()

    cost_volume = cost_volume/(reweight_sum+1e-10)

    return cost_volume, vis_weight_list

#   def homography_warp_coords(cfg, features, level, ref_in, src_in, ref_pose, src_pose, depth_hypos, coords, H, W, gwc_groups, vwa_net=None, vis_weights=None):
#       batch_size, c, h, w = features[0][level].shape
#       _, _, num_planes, _, _ = depth_hypos.shape
#       _, num_pixels, _ = coords.shape
#       num_src_views = len(features)-1
#   
#       K_ref = torch.zeros(batch_size, 4, 4).to(ref_in)
#       K_ref[:, :3,:3] = ref_in
#       K_ref[:, 3, 3] = 1
#   
#       K_src = torch.zeros(batch_size, num_src_views, 4, 4).to(ref_in)
#       for v in range(num_src_views):
#           K_src[:, v, :3, :3] = src_in[:,v]
#       K_src[:, :, 3, 3] = 1
#   
#       vis_weight_list = []
#       cost_volume = None
#       reweight_sum = None
#   
#       # build coordinates vector
#       depth_hypos = depth_hypos.reshape(batch_size, num_planes, num_pixels) # batch_size x num_planes x num_pixels
#       coords = torch.movedim(coords, (0,1,2), (0,2,1)).to(torch.float32) # batch_size x 2 x num_pixels
#       x_coords = coords[:,1,:]
#       y_coords = coords[:,0,:]
#       xyz = torch.stack((x_coords, y_coords, torch.ones_like(x_coords)), dim=1) # batch_size, 3 x num_pixels
#   
#       # sample reference features
#       x_normalized = ((x_coords / (W-1)) * 2) - 1
#       y_normalized = ((y_coords / (H-1)) * 2) - 1
#       xy = torch.stack((x_normalized, y_normalized), dim=-1)  # [B, num_pixels, 2]
#       ref_features = features[0][level]
#       ref_features = F.grid_sample(ref_features,
#                               xy.view(batch_size, 1, num_pixels, 2),
#                               mode='nearest',
#                               padding_mode='zeros',
#                               align_corners=False)
#       ref_features = ref_features.repeat(1,1,num_planes,1) # [B x C x num_pixels x num_planes]
#   
#       for src in range(num_src_views):
#           with torch.no_grad():
#               src_proj = torch.matmul(K_src[:,src],src_pose[:,src])
#               ref_proj = torch.matmul(K_ref,ref_pose)
#   
#               proj = torch.matmul(src_proj, torch.inverse(ref_proj))
#               rot = proj[:, :3, :3]  # [B,3,3]
#               trans = proj[:, :3, 3:4]  # [B,3,1]
#   
#               rot_xyz = torch.matmul(rot, xyz)  # [B, 3, num_pixels]
#               rot_depth_xyz = rot_xyz.unsqueeze(2).repeat(1, 1, num_planes, 1) * depth_hypos.view(batch_size, 1, num_planes, num_pixels)  # [B, 3, num_planes, num_pixels]
#               proj_xyz = rot_depth_xyz + trans.view(batch_size, 3, 1, 1)  # [B, 3, num_planes, num_pixels]
#               proj_xy = proj_xyz[:, :2, :, :] / proj_xyz[:, 2:3, :, :]  # [B, 2, num_planes, num_pixels]
#               proj_x_normalized = ((proj_xy[:, 0, :, :] / (W-1)) * 2) - 1
#               proj_y_normalized = ((proj_xy[:, 1, :, :] / (H-1)) * 2) - 1
#   
#               proj_xy = torch.stack((proj_x_normalized, proj_y_normalized), dim=3)  # [B, num_planes, num_pixels, 2]
#               grid = proj_xy
#   
#           grid = grid.type(ref_features.dtype)
#           # Note: The shape of the input map is 4D: Batch x Channels x Height x Width.
#           #       The input coordinates must be in [x, y] format where x->width, y->height.
#           #       These coordinates must be normalized between [-1, 1].
#           src_features = features[src+1][level]
#           src_features = F.grid_sample(src_features,
#                                   grid.view(batch_size, num_planes, num_pixels, 2),
#                                   mode='bilinear',
#                                   padding_mode='zeros',
#                                   align_corners=False)
#           two_view_cost_volume = groupwise_correlation(src_features, ref_features, gwc_groups[level]) #B,C,num_planes,num_pixels
#           two_view_cost_volume = two_view_cost_volume.reshape(batch_size, gwc_groups[level], num_planes, H, W)
#   
#           # Estimate visability weight for init level
#           if vwa_net is not None:
#               reweight = vwa_net(two_view_cost_volume) #B, H, W
#               vis_weight_list.append(reweight)
#               reweight = reweight.unsqueeze(1).unsqueeze(2) #B, 1, 1, H, W
#               two_view_cost_volume = reweight*two_view_cost_volume
#   
#           # Use estimated visability weights for refine levels
#           elif vis_weights is not None:
#               reweight = vis_weights[src].unsqueeze(1)
#               if reweight.shape[2] < two_view_cost_volume.shape[3]:
#                   reweight = F.interpolate(reweight,scale_factor=2,mode='bilinear',align_corners=False)
#               vis_weight_list.append(reweight.squeeze(1))
#               reweight = reweight.unsqueeze(2)
#               two_view_cost_volume = reweight*two_view_cost_volume
#   
#           if cost_volume == None:
#               cost_volume = two_view_cost_volume
#               reweight_sum = reweight
#           else:
#               cost_volume = cost_volume + two_view_cost_volume
#               reweight_sum = reweight_sum + reweight
#   
#       cost_volume = cost_volume/(reweight_sum+1e-5)
#       return cost_volume, vis_weight_list


def match_features(src_image: np.ndarray, tgt_image: np.ndarray, max_features: int = 500) -> Tuple[np.ndarray, np.ndarray]:
    """Computer matching ORB features between a pair of images.

    Args:
        src_image: The source image to compute and match features.
        tgt_image: The target image to compute and match features.
        max_features: The maximum number of features to retain.

    Returns:
        src_points: The set of matched point coordinates for the source image.
        tgt_points: The set of matched point coordinates for the target image.
    """
    src_image = cv2.cvtColor(src_image,cv2.COLOR_BGR2GRAY)
    tgt_image = cv2.cvtColor(tgt_image, cv2.COLOR_BGR2GRAY)
      
    orb = cv2.ORB_create(max_features)
      
    src_keypoints, src_descriptors = orb.detectAndCompute(src_image,None)
    tgt_keypoints, tgt_descriptors = orb.detectAndCompute(tgt_image,None)

    matcher = cv2.BFMatcher(crossCheck=True)
    matches = list(matcher.match(src_descriptors, tgt_descriptors) )
    matches.sort(key = lambda x:x.distance)

    src_points = []
    tgt_points = []
    for i in range(8):
        m = matches[i]

        src_points.append(src_keypoints[m.queryIdx].pt)
        tgt_points.append(tgt_keypoints[m.trainIdx].pt)
    src_points  = np.asarray(src_points)
    tgt_points = np.asarray(tgt_points)

    return (src_points, tgt_points)

def plane_coords(K, P, depth_hypos, H, W):
    """Batched PyTorch version
    """
    batch_size,_,_ = K.shape
    num_planes = depth_hypos.shape[0]

    xyz = torch.movedim(torch.tensor([[0,0,1], [W-1,0,1], [0,H-1,1], [W-1,H-1,1]], dtype=torch.float32), 0, 1).to(P)
    xyz = xyz.reshape(1,3,4).repeat(batch_size, 1, 1)
    if K.shape[1]==3:
        K_44 = torch.zeros((batch_size, 4, 4)).to(P)
        K_44[:,:3,:3] = K[:,:3,:3]
        K_44[:,3,3] = 1
        K = K_44
    proj = K @ P
    inv_proj = torch.linalg.inv(proj)

    planes = torch.zeros(num_planes, 3, 4).to(inv_proj)
    for p in range(num_planes):
        planes[p] = (inv_proj[0,:3,:3] @ xyz) * depth_hypos[p]
        planes[p] += inv_proj[0,:3,3:4]

    return planes


def _plane_coords(K, P, near, far, H, W):
    """Numpy version
    """
    xyz = np.asarray([[0,0,1], [W-1,0,1], [0,H-1,1], [W-1,H-1,1]], dtype=np.float32).transpose()
    if K.shape[0]==3:
        K_44 = np.zeros((4, 4))
        K_44[:3,:3] = K[:3,:3]
        K_44[3,3] = 1
        K = K_44
    proj = K @ P

    near_plane = (np.linalg.inv(proj)[:3,:3] @ xyz) * near
    near_plane += np.linalg.inv(proj)[:3,3:4]
    far_plane = (np.linalg.inv(proj)[:3,:3] @ xyz) * far
    far_plane += np.linalg.inv(proj)[:3,3:4]

    return near_plane, far_plane

def _points_from_depth(depth: np.ndarray, intrinsics: np.ndarray, extrinsics: np.ndarray) -> np.ndarray:
    """Creates a point array from a single depth map.

    Parameters:
        depth: Depth map to project to 3D.
        intrinsics: Intrinsic camera parameters for the given depth map viewpoint.
        extrinsics: Extrinsic camera parameters for the given depth map viewpoint.

    Returns:
        An array of 3D points corresponding to the input depth map.
    """
    # project depth map to point cloud
    height, width = depth.shape
    x = np.linspace(0,width-1,num=width)
    y = np.linspace(0,height-1,num=height)
    x,y = np.meshgrid(x,y, indexing="xy")
    x = x.flatten()
    y = y.flatten()
    depth = depth.flatten()
    valid_inds = np.argwhere(depth>0)[:,0]
    xyz_cam = np.matmul(np.linalg.inv(intrinsics[:3,:3]), np.vstack((x, y, np.ones_like(x))) * depth)
    xyz_world = np.matmul(np.linalg.inv(extrinsics), np.vstack((xyz_cam, np.ones_like(x))))[:3]
    points = xyz_world.transpose((1, 0))
    return points, valid_inds

def project_depth_map(depth: torch.Tensor, cam: torch.Tensor, mask: Optional[torch.Tensor] = None) -> torch.Tensor:
    """Projects a depth map into a list of 3D points

    Parameters:
        depth: Input depth map to project.
        cam: Camera parameters for input depth map.

    Returns:
        A float Tensor of 3D points corresponding to the projected depth values.
    """
    if (depth.shape[1] == 1):
        depth = depth.squeeze(1)

    batch_size, height, width = depth.shape
    cam_shape = cam.shape

    # get camera extrinsics and intrinsics
    P = cam[:,0,:,:]
    K = cam[:,1,:,:]
    K[:,3,:] = torch.tensor([0,0,0,1])

    # construct back-projection from invers matrices
    # separate into rotation and translation components
    bwd_projection = torch.matmul(torch.inverse(P), torch.inverse(K)).to(torch.float32)
    bwd_rotation = bwd_projection[:,:3,:3]
    bwd_translation = bwd_projection[:,:3,3:4]

    # build 2D homogeneous coordinates tensor: [B, 3, H*W]
    with torch.no_grad():
        row_span = torch.arange(0, height, dtype=torch.float32).cuda()
        col_span = torch.arange(0, width, dtype=torch.float32).cuda()
        r,c = torch.meshgrid(row_span, col_span, indexing="ij")
        r,c = r.contiguous(), c.contiguous()
        r,c = r.reshape(height*width), c.reshape(height*width)
        coords = torch.stack((c,r,torch.ones_like(c)))
        coords = torch.unsqueeze(coords, dim=0).repeat(batch_size, 1, 1)

    # compute 3D coordinates using the depth map: [B, H*W, 3]
    world_coords = torch.matmul(bwd_rotation, coords)
    depth = depth.reshape(batch_size, 1, -1)
    world_coords = world_coords * depth
    world_coords = world_coords + bwd_translation

    #TODO: make sure index select is differentiable
    #       (there is a backward function but need to find the code..)
    if (mask != None):
        world_coords = torch.index_select(world_coords, dim=2, index=non_zero_inds)
        world_coords = torch.movedim(world_coords, 1, 2)

    # reshape 3D coordinates back into 2D map: [B, H, W, 3]
    #   coords_map = world_coords.reshape(batch_size, height, width, 3)

    return world_coords

def project_renderer(renderer: o3d.visualization.rendering.OffscreenRenderer, K: np.ndarray, P: np.ndarray, width: float, height: float) -> np.ndarray:
    """Projects the scene in an Open3D Offscreen Renderer to the 2D image plane.

    Parameters:
        renderer: Geometric scene to be projected.
        K: Camera intrinsic parameters.
        P: Camera extrinsic parameters.
        width: Desired image width.
        height: Desired image height.

    Returns:
        The rendered image for the scene at the specified camera viewpoint.
    """
    # set up the renderer
    intrins = o3d.camera.PinholeCameraIntrinsic(width, height, K[0,0], K[1,1], K[0,2], K[1,2])
    renderer.setup_camera(intrins, P)

    # render image
    image = np.asarray(renderer.render_to_image())
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

    return image

def render_custom_values(points: np.ndarray, values: np.ndarray, image_shape: Tuple[int,int], cam: np.ndarray) -> np.ndarray:
    """Renders a point cloud into a 2D camera plane using custom values for each pixel.

    Parameters:
        points: List of 3D points to be rendered.
        values: List of values to be written in the rendered image.
        image_shape: Desired shape (height,width) of the rendered image.
        cam: Camera parameters for the image viewpoint.

    Returns:
        The rendered image for the list of points using the sepcified corresponding values.
    """
    points = points.tolist()
    values = list(values.astype(float))
    cam = cam.flatten().tolist()

    rendered_img = rd.render(list(image_shape), points, values, cam)

    return rendered_img

def _render_point_cloud(cloud: o3d.geometry.PointCloud, pose: np.ndarray, K: np.ndarray, width: int, height: int) -> np.ndarray:
    """Renders a point cloud into a 2D image plane.

    Parameters:
        cloud: Point cloud to be rendered.
        pose: Camera extrinsic parameters for the image plane.
        K: Camera intrinsic parameters for the image plane.
        width: Desired width of the rendered image.
        height: Desired height of the rendered image.

    Returns:
        The rendered image for the point cloud at the specified camera viewpoint.
    """
    #   cmap = plt.get_cmap("hot_r")
    #   colors = cmap(dists)[:, :3]
    #   ply.colors = o3d.utility.Vector3dVector(colors)

    # set up the renderer
    render = o3d.visualization.rendering.OffscreenRenderer(width, height)
    mat = o3d.visualization.rendering.MaterialRecord()
    mat.shader = 'defaultUnlit'
    render.scene.add_geometry("cloud", cloud, mat)
    render.scene.set_background(np.asarray([0,0,0,1])) #r,g,b,a
    intrins = o3d.camera.PinholeCameraIntrinsic(width, height, K[0,0], K[1,1], K[0,2], K[1,2])
    render.setup_camera(intrins, pose)

    # render image
    image = np.asarray(render.render_to_image())
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

    depth = np.asarray(render.render_to_depth_image(z_in_view_space=True))

    return image, depth

def render_point_cloud_single(cloud: o3d.geometry.PointCloud, pose: np.ndarray, K: np.ndarray, width: int, height: int) -> np.ndarray:
    """Renders a point cloud into a 2D image plane.

    Parameters:
        cloud: Point cloud to be rendered.
        pose: Camera extrinsic parameters for the image plane.
        K: Camera intrinsic parameters for the image plane.
        width: Desired width of the rendered image.
        height: Desired height of the rendered image.

    Returns:
        The rendered image for the point cloud at the specified camera viewpoint.
    """
    #   cmap = plt.get_cmap("hot_r")
    #   colors = cmap(dists)[:, :3]
    #   ply.colors = o3d.utility.Vector3dVector(colors)

    # set up the renderer
    render = o3d.visualization.rendering.OffscreenRenderer(width, height)
    mat = o3d.visualization.rendering.MaterialRecord()
    mat.shader = 'defaultUnlit'
    render.scene.add_geometry("cloud", cloud, mat)
    render.scene.set_background(np.asarray([0,0,0,1])) #r,g,b,a
    intrins = o3d.camera.PinholeCameraIntrinsic(width, height, K[0,0], K[1,1], K[0,2], K[1,2])
    render.setup_camera(intrins, pose)

    # render image
    image = np.asarray(render.render_to_image())
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

    depth = np.asarray(render.render_to_depth_image(z_in_view_space=True))

    return image, depth

def render_point_cloud(render, intrins, pose):
    """Renders a point cloud into a 2D image plane.

    Parameters:

    Returns:
    """
    render.setup_camera(intrins, pose)

    # render image
    image = np.asarray(render.render_to_image())
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

    depth = np.asarray(render.render_to_depth_image(z_in_view_space=True))

    return image, depth

def reproject(src_depth, src_K, src_P, tgt_depth, tgt_K, tgt_P):
    """Computes the re-projection depth values and pixel indices between two depth maps.

    This function takes as input two depth maps: 'src_depth' and 'tgt_depth'. The source
    depth map is first projected into the target camera plane using the source depth
    values and the camera parameters for both views. Using the projected pixel
    coordinates in the target view, the target depths are then re-projected back into
    the source camera plane (again with the camera parameters for both views). The
    information prouced from this process is often used to compute errors in
    re-projection between two depth maps, or similar operations.

    Parameters:
        src_depth: Source depth map to be projected.
        src_K: Intrinsic camera parameters for the source depth map viewpoint.
        src_P: Extrinsic camera parameters for the source depth map viewpoint.
        tgt_depth: Target depth map used for re-projection.
        tgt_K: Intrinsic camera parameters for the target depth map viewpoint.
        tgt_P: Extrinsic camera parameters for the target depth map viewpoint.

    Returns:
        depth_reprojected: The re-projected depth values for the source depth map.
        coords_reprojected: The re-projection coordinates for the source view.
        coords_tgt: The projected coordinates for the target view.
    """
    batch_size, c, height, width = src_depth.shape

    # back-project ref depths to 3D
    x_src, y_src = torch.meshgrid(torch.arange(0, width), torch.arange(0, height), indexing="xy")
    x_src = x_src.reshape(-1).unsqueeze(0).repeat(batch_size, 1).to(src_depth)
    y_src = y_src.reshape(-1).unsqueeze(0).repeat(batch_size, 1).to(src_depth)
    homog = torch.stack((x_src, y_src, torch.ones_like(x_src)), dim=1)
    xyz_src = torch.matmul(torch.linalg.inv(src_K), homog * src_depth.reshape(batch_size, 1, -1))

    # transform 3D points from ref to src coords
    homog_3d = torch.concatenate((xyz_src, torch.ones_like(x_src).unsqueeze(1)), dim=1)
    xyz_tgt = torch.matmul(torch.matmul(tgt_P, torch.linalg.inv(src_P)), homog_3d)[:,:3]

    # project src 3D points into pixel coords
    K_xyz_tgt = torch.matmul(tgt_K, xyz_tgt)
    xy_tgt = K_xyz_tgt[:,:2] / K_xyz_tgt[:,2:3]
    x_tgt = xy_tgt[:,0].reshape(batch_size, height, width).to(torch.float32)
    y_tgt = xy_tgt[:,1].reshape(batch_size, height, width).to(torch.float32)
    coords_tgt = torch.stack((x_tgt, y_tgt), dim=-1) # B x H x W x 2

    # sample the depth values from the src map at each pixel coord
    x_normalized = ((x_tgt / (width-1)) * 2) - 1
    y_normalized = ((y_tgt / (height-1)) * 2) - 1
    grid = torch.stack((x_normalized, y_normalized), dim=-1) # B x H x W x 2
    sampled_depth_tgt = F.grid_sample(
                                    tgt_depth,
                                    grid,
                                    mode="bilinear",
                                    padding_mode="zeros",
                                    align_corners=False)

    # back-project src depths to 3D
    homog = torch.concatenate((xy_tgt, torch.ones_like(x_src).unsqueeze(1)), dim=1)
    xyz_tgt = torch.matmul(torch.linalg.inv(tgt_K), homog * sampled_depth_tgt.reshape(batch_size, 1, -1))

    # transform 3D points from src to ref coords
    homog_3d = torch.concatenate((xyz_tgt, torch.ones_like(x_src).unsqueeze(1)), dim=1)
    xyz_reprojected = torch.matmul(torch.matmul(src_P, torch.linalg.inv(tgt_P)), homog_3d)[:,:3]

    # extract reprojected depth values
    depth_reprojected = xyz_reprojected[:,2].reshape(batch_size, height, width).to(torch.float32)

    # project ref 3D points into pixel coords
    K_xyz_reprojected = torch.matmul(src_K, xyz_reprojected)
    xy_reprojected = K_xyz_reprojected[:,:2] / (K_xyz_reprojected[:,2:3] + 1e-7)
    x_reprojected = xy_reprojected[:,0].reshape(batch_size, height, width).to(torch.float32)
    y_reprojected = xy_reprojected[:,1].reshape(batch_size, height, width).to(torch.float32)

    coords_reprojected = torch.stack((x_reprojected, y_reprojected), dim=-1) # B x H x W x 2

    return depth_reprojected, coords_reprojected, coords_tgt

def _reproject(src_depth, src_K, src_P, tgt_depth, tgt_K, tgt_P):
    """Computes the re-projection depth values and pixel indices between two depth maps.

    This function takes as input two depth maps: 'src_depth' and 'tgt_depth'. The source
    depth map is first projected into the target camera plane using the source depth
    values and the camera parameters for both views. Using the projected pixel
    coordinates in the target view, the target depths are then re-projected back into
    the source camera plane (again with the camera parameters for both views). The
    information prouced from this process is often used to compute errors in
    re-projection between two depth maps, or similar operations.

    Parameters:
        src_depth: Source depth map to be projected.
        src_K: Intrinsic camera parameters for the source depth map viewpoint.
        src_P: Extrinsic camera parameters for the source depth map viewpoint.
        tgt_depth: Target depth map used for re-projection.
        tgt_K: Intrinsic camera parameters for the target depth map viewpoint.
        tgt_P: Extrinsic camera parameters for the target depth map viewpoint.

    Returns:
        depth_reprojected: The re-projected depth values for the source depth map.
        coords_reprojected: The re-projection coordinates for the source view.
        coords_tgt: The projected coordinates for the target view.
    """
    height, width = src_depth.shape

    # back-project ref depths to 3D
    x_src, y_src = np.meshgrid(np.arange(0, width), np.arange(0, height))
    x_src, y_src = x_src.reshape([-1]), y_src.reshape([-1])
    xyz_src = np.matmul(np.linalg.inv(src_K),
                        np.vstack((x_src, y_src, np.ones_like(x_src))) * src_depth.reshape([-1]))

    # transform 3D points from ref to src coords
    xyz_tgt = np.matmul(np.matmul(tgt_P, np.linalg.inv(src_P)),
                        np.vstack((xyz_src, np.ones_like(x_src))))[:3]

    # project src 3D points into pixel coords
    K_xyz_tgt = np.matmul(tgt_K, xyz_tgt)
    xy_tgt = K_xyz_tgt[:2] / K_xyz_tgt[2:3]
    x_tgt = xy_tgt[0].reshape([height, width]).astype(np.float32)
    y_tgt = xy_tgt[1].reshape([height, width]).astype(np.float32)

    # sample the depth values from the src map at each pixel coord
    sampled_depth_tgt = cv2.remap(tgt_depth, x_tgt, y_tgt, interpolation=cv2.INTER_LINEAR)

    # back-project src depths to 3D
    xyz_tgt = np.matmul(np.linalg.inv(tgt_K),
                        np.vstack((xy_tgt, np.ones_like(x_src))) * sampled_depth_tgt.reshape([-1]))

    # transform 3D points from src to ref coords
    xyz_reprojected = np.matmul(np.matmul(src_P, np.linalg.inv(tgt_P)),
                                np.vstack((xyz_tgt, np.ones_like(x_src))))[:3]

    # extract reprojected depth values
    depth_reprojected = xyz_reprojected[2].reshape([height, width]).astype(np.float32)

    # project ref 3D points into pixel coords
    K_xyz_reprojected = np.matmul(src_K, xyz_reprojected)
    xy_reprojected = K_xyz_reprojected[:2] / (K_xyz_reprojected[2:3] + 1e-7)
    x_reprojected = xy_reprojected[0].reshape([height, width]).astype(np.float32)
    y_reprojected = xy_reprojected[1].reshape([height, width]).astype(np.float32)

    coords_reprojected = np.dstack((x_reprojected, y_reprojected))
    coords_tgt = np.dstack((x_tgt, y_tgt))

    return depth_reprojected, coords_reprojected, coords_tgt

def sample_volume(volume, z_vals, coords, H, W, near_depth, far_depth, inv_depth):
    """
    Parameters:

    Returns:
    """
    N, M = z_vals.shape
    batch_size, c, _, _, _ = volume.shape

    z_vals = z_vals.reshape(N,M,1) # N x M x 1
    if inv_depth:
        z_vals = 1/z_vals
        near_depth = 1/near_depth
        far_depth = 1/far_depth
    coords = coords.reshape(N,1,2).repeat(1,M,1) # N x M x 2
    x_coords = coords[:,:,1:2]
    y_coords = coords[:,:,0:1]
    points = torch.cat([x_coords, y_coords, z_vals], dim=-1) # N x M x 3
    points = torch.reshape(points, [-1, 3]) # N*M x 3

    # define coordinates bounds
    min_coord = torch.tensor([0,0,near_depth]).to(points)
    max_coord = torch.tensor([W-1,H-1,far_depth]).to(points)
    min_coord = min_coord.reshape(1,3).repeat(N*M,1)
    max_coord = max_coord.reshape(1,3).repeat(N*M,1) 

    # normalize points
    norm_points = (points - min_coord) / (max_coord - min_coord)
    norm_points = norm_points.unsqueeze(0).repeat(batch_size, 1, 1)
    norm_points = (norm_points * 2) - 1

    # Note: The shape of the input volume is 5D: Batch x Channels x Depth x Height x Width.
    #       The input coordinates must be in [x, y, z] format where x->width, y->height, z->depth.
    #       These coordinates must be normalized between [-1, 1].
    features = F.grid_sample(volume,
                            norm_points.view(batch_size, N*M, 1, 1, 3),
                            mode='bilinear',
                            padding_mode='zeros',
                            align_corners=True)
    features = torch.movedim(features.reshape(c, N*M), 0, 1) # N*M x c

    return features

def soft_hypothesis(data, target_hypo, focal_length, min_hypo, max_hypo, M, delta_in=1):
    """
    Parameters:

    Returns:
    """
    B, _, D, H, W = target_hypo.shape
    rand_match_offset = torch.rand(B,1,M,H,W).to(target_hypo)
    near, far = Z_from_disp(target_hypo, data["baseline"], focal_length, delta=delta_in)
    target_range = torch.abs(far - near).repeat(1,1,M,1,1)

    target_samples = (rand_match_offset * target_range) + near
    mask = torch.where(target_hypo <= 0, 0.0, 1.0).repeat(1,1,M,1,1)
    matching_hypos = torch.clip(target_samples, min_hypo, max_hypo) * mask

    return matching_hypos

def _soft_hypothesis(data, target_hypo, focal_length, min_hypo, max_hypo, M=1, delta_in=1):
    """
    Parameters:

    Returns:
    """
    B, _, D, H, W = target_hypo.shape
    rand_match_offset = torch.rand(B,1,M,H,W).to(target_hypo)
    rand_near_offset = torch.rand(B,1,M,H,W).to(target_hypo)
    rand_far_offset = torch.rand(B,1,M,H,W).to(target_hypo)

    near, far = Z_from_disp(target_hypo, data["baseline"], focal_length, delta=delta_in)
    target_range = torch.abs(far - near).repeat(1,1,M,1,1)
    near_range = torch.abs(near - min_hypo).repeat(1,1,M,1,1)
    far_range = torch.abs(max_hypo - far).repeat(1,1,M,1,1)

    target_samples = (rand_match_offset * target_range) + near
    near_samples = (rand_near_offset * near_range) + min_hypo
    far_samples = (rand_far_offset * far_range) + far
    samples = torch.cat([target_samples,near_samples,far_samples], dim=1)
    samples = samples.reshape(B,-1,H,W).unsqueeze(1) # [B, 1, M*3, H, W]

    mask = torch.where(target_hypo <= 0, 0.0, 1.0).repeat(1,1,M*3,1,1)
    hypos = torch.clip(samples, min_hypo, max_hypo) * mask

    return hypos

def resolution_based_hypothesis(data, target_hypo, level, focal_length, min_hypo, max_hypo, delta_in=1):
    """
    Parameters:

    Returns:
    """
    B, _, D, H, W = target_hypo.shape
    rand_match_offset = torch.rand(B,1,M,H,W).to(target_hypo)
    rand_near_offset = torch.rand(B,1,M,H,W).to(target_hypo)
    rand_far_offset = torch.rand(B,1,M,H,W).to(target_hypo)

    near, far = Z_from_disp(target_hypo, data["baseline"], focal_length, delta=delta_in)
    target_range = torch.abs(far - near).repeat(1,1,M,1,1)
    near_range = torch.abs(near - min_hypo).repeat(1,1,M,1,1)
    far_range = torch.abs(max_hypo - far).repeat(1,1,M,1,1)

    target_samples = (rand_match_offset * target_range) + near
    near_samples = (rand_near_offset * near_range) + min_hypo
    far_samples = (rand_far_offset * far_range) + far
    samples = torch.cat([target_samples,near_samples,far_samples], dim=1)
    samples = samples.reshape(B,-1,H,W).unsqueeze(1) # [B, 1, M*3, H, W]

    mask = torch.where(target_hypo <= 0, 0.0, 1.0).repeat(1,1,M*3,1,1)
    hypos = torch.clip(samples, min_hypo, max_hypo) * mask

    return hypos

def visibility(depths, K, Ps, vis_th, levels=4):
    """
    Parameters:

    Returns:
    """
    batch_size, views, c, H, W = depths.shape

    K_pyr = intrinsic_pyramid(K, levels)

    vis_maps = []
    vis_masks = []
    for l in range(levels):
        resized_depths = tvf.resize(depths[:,:,0], [int(H/(2**l)), int(W/(2**l))]).unsqueeze(2)
        batch_size, views, c, height, width = resized_depths.shape
        vis_map = torch.where(resized_depths[:,0] > 0.0, 1, 0)

        for i in range(1, views):
            mask = geometric_consistency_mask(resized_depths[:,0], K_pyr[:,l], Ps[:,0], resized_depths[:,i], K_pyr[:,l], Ps[:,i], pixel_th=0.5)
            vis_map += mask.unsqueeze(1)
        vis_map = vis_map.to(torch.float32)

        vis_maps.append(vis_map)
        vis_masks.append(torch.where(vis_map >= vis_th, 1, 0))
    return vis_maps, vis_masks

def _visibility(depths, K, Ps, vis_th=None):
    """
    Parameters:

    Returns:
    """
    views, height, width = depths.shape
    vis_map = np.where(depths[0] > 0.0, 1, 0)

    for i in range(1, views):
        mask = _geometric_consistency_mask(depths[0], K, Ps[0], depths[i], K, Ps[i], pixel_th=0.75)
        vis_map += mask
    vis_map = vis_map.astype(np.float32)

    if vis_th != None:
        vis_map = np.where(vis_map >= vis_th, 1, 0)

    return vis_map

def visibility_mask(src_depth: np.ndarray, src_cam: np.ndarray, depth_files: List[str], cam_files: List[str], src_ind: int = -1, pixel_th: float = 0.1) -> np.ndarray:
    """Computes a visibility mask between a provided source depth map and list of target depth maps.

    Parameters:
        src_depth: Depth map for the source view.
        src_cam: Camera parameters for the source depth map viewpoint.
        depth_files: List of target depth maps.
        cam_files: List of corresponding target camera parameters for each targte depth map viewpoint.
        src_ind: Index into 'depth_files' corresponding to the source depth map (if included in the list).
        pixel_th: Pixel re-projection threshold to determine matching depth estimates.

    Returns:
        The visibility mask for the source view.
    """
    height, width = src_depth.shape
    vis_map = np.not_equal(src_depth, 0.0).astype(np.double)

    for i in range(len(depth_files)):
        if (i==src_ind):
            continue

        # get files
        sdf = depth_files[i]
        scf = cam_files[i]

        # load data
        tgt_depth = read_pfm(sdf)
        tgt_cam = read_single_cam_sfm(scf,'r')

        mask = geometric_consistency_mask(src_depth, src_cam, tgt_depth, tgt_cam, pixel_th)
        vis_map += mask

    return vis_map.astype(np.float32)

def uniform_hypothesis(cfg, device, batch_size, depth_min, depth_max, img_height, img_width, planes, inv_depth=False, bin_format=False):
    """
    Parameters:

    Returns:
    """
    depth_range = depth_max-depth_min

    hypotheses = torch.zeros((batch_size,planes),device=device)
    for b in range(0,batch_size):
        if bin_format:
            spacing = depth_range/planes
            start_depth = depth_min + (spacing/2)
            end_depth = depth_min + (spacing/2) + ((planes-1)*spacing)
        else:
            start_depth = depth_min
            end_depth = depth_max
        if inv_depth:
            hypotheses[b] = 1/(torch.linspace(1/start_depth,1/end_depth,steps=planes,device=device))
        else:
            hypotheses[b] = torch.linspace(start_depth, end_depth, steps=planes,device=device)
    hypotheses = hypotheses.unsqueeze(2).unsqueeze(3).repeat(1,1,img_height,img_width)

    # Make coordinate for depth hypothesis, to be used by sparse convolution.
    depth_hypo_coords = torch.zeros((batch_size,planes),device=device)
    for b in range(0,batch_size):
        depth_hypo_coords[b] = torch.linspace(0,planes-1,steps=planes,device=device)
    depth_hypo_coords = depth_hypo_coords.unsqueeze(2).unsqueeze(3).repeat(1,1,img_height,img_width)

    # Calculate hypothesis interval
    hypo_intervals = hypotheses[:,1:]-hypotheses[:,:-1]
    hypo_intervals = torch.cat((hypo_intervals,hypo_intervals[:,-1].unsqueeze(1)),dim=1)

    return hypotheses.unsqueeze(1), depth_hypo_coords.unsqueeze(1), hypo_intervals.unsqueeze(1)
